"""Custom exceptions for the ``osbng`` package.

These exceptions are intended to provide clear and specific error handling for
scenarios where invalid inputs or operations are encountered.

Custom exceptions:

- **BNGExtentError**: Raised when easting and northing coordinates fall outside of the
  BNG index system extent.
- **BNGHierarchyError**: Raised when an invalid parent or child derivation is attempted.
- **BNGNeighbourError**: Raised when an invalid neighbour relationship is encountered.
- **BNGReferenceError**: Raised when an invalid BNG reference string is provided.
- **BNGResolutionError**: Raised when an invalid BNG resolution is provided.
"""

from osbng.resolution import BNG_RESOLUTIONS


class BNGReferenceError(Exception):
    """Raised for invalid BNG reference strings."""

    pass


class BNGResolutionError(Exception):
    """Raised for unsupported BNG resolutions."""

    def __init__(self):
        """Initialise exception with a message listing supported resolutions."""
        # Extract the numeric and string resolutions from BNG_RESOLUTIONS
        # Create message listing supported resolutions
        message = (
            "Invalid BNG resolution provided. Supported resolutions are: \n"
            + f"Metres: {', '.join(map(str, BNG_RESOLUTIONS.keys()))}\n"
            + "Labels: "
            + f"{', '.join(value['label'] for value in BNG_RESOLUTIONS.values())}"
        )
        # Pass message to base class
        super().__init__(message)


class BNGHierarchyError(Exception):
    """Raised for invalid parent/child derivation."""

    pass


class BNGNeighbourError(Exception):
    """Raised for invalid neighbour relationships."""

    pass


class BNGExtentError(Exception):
    """Raised for coordinates outside the BNG index system extent.

    BNG extent defined as 0 <= easting < 700000 and 0 <= northing < 1300000
    """

    def __init__(self):
        """Initialise exception with a message listing the valid coordinate ranges."""
        # Create message listing the easting and northing coordinate ranges
        message = (
            "Coordinates outside of the BNG extent. "
            "Easting and northing values must be within: \n"
            "0 <= easting < 700000\n"
            "0 <= northing < 1300000"
        )
        # Pass message to base class
        super().__init__(message)


# Map exception strings to exception classes
_EXCEPTION_MAP = {
    "BNGReferenceError": BNGReferenceError,
    "BNGResolutionError": BNGResolutionError,
    "BNGHierarchyError": BNGHierarchyError,
    "BNGNeighbourError": BNGNeighbourError,
    "BNGExtentError": BNGExtentError,
}



"""Testing for the bng_reference module.

The test cases are defined in the JSON file located at
./data/bng_reference_test_cases.json and are used to parameterise the tests for various
functions in the indexing module.

Test cases are loaded from the JSON file using the _load_test_cases function, which is
defined in the utils module. The test cases are defined as TypedDicts, which provide a
way to define the structure of the test case data.
"""

from pathlib import Path
from typing import TypedDict

import pytest

from osbng.bng_reference import (
    BNGReference,
    _format_bng_ref_string,
    _get_bng_resolution_label,
    _get_bng_resolution_metres,
    _validate_bng_ref_string,
)
from osbng.errors import _EXCEPTION_MAP
from osbng.utils import _load_test_cases


class ValidateBNGRefStringTestCase(TypedDict):
    """TypedDict for _validate_bng_ref_string function test cases.

    Attributes:
        bng_ref_string (str): The BNG reference string to validate.
        expected (bool): True if the BNG reference is valid, False otherwise.
    """

    bng_ref_string: str
    expected: bool


# Parameterised test for _validate_bng_ref_string function
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "bng_reference_test_cases.json"
    )["_validate_bng_ref_string"],
)
def test__validate_bng_ref_string(test_case: ValidateBNGRefStringTestCase) -> None:
    """Test _validate_bng_ref_string function with test cases from JSON file.

    Args:
        test_case (ValidateBNGRefStringTestCase): Test case from JSON file.
    """
    # Load test case data
    bng_ref_string = test_case["bng_ref_string"]
    expected = test_case["expected"]
    # Assert that the function returns the expected result
    assert _validate_bng_ref_string(bng_ref_string) == expected


class GetBNGResolutionMetresTestCase(TypedDict):
    """TypedDict for _get_bng_resolution_metres function test cases.

    Attributes:
        bng_ref_string (str): The BNG reference string to validate.
        expected (int): The resolution expressed as a metre-based integer.
    """

    bng_ref_string: str
    expected: int


# Parameterised test for _get_bng_resolution_metres function
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "bng_reference_test_cases.json"
    )["_get_bng_resolution_metres"],
)
def test__get_bng_resolution_metres(test_case: GetBNGResolutionMetresTestCase) -> None:
    """Test _get_bng_resolution_metres function with test cases from JSON file.

    Args:
        test_case (GetBNGResolutionMetresTestCase): Test case from JSON file.
    """
    # Load test case data
    bng_ref_string = test_case["bng_ref_string"]
    expected = test_case["expected"]
    # Assert that the function returns the expected result
    assert _get_bng_resolution_metres(bng_ref_string) == expected


class GetBNGResolutionLabelTestCase(TypedDict):
    """TypedDict for _get_bng_resolution_label function test cases.

    Attributes:
        bng_ref_string (str): The BNG reference string to validate.
        expected (str): The resolution expressed as a descriptive string.
    """

    bng_ref_string: str
    expected: str


# Parameterised test for _get_bng_resolution_label function
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "bng_reference_test_cases.json"
    )["_get_bng_resolution_label"],
)
def test__get_bng_resolution_label(test_case: GetBNGResolutionLabelTestCase) -> None:
    """Test _get_bng_resolution_label function with test cases from JSON file.

    Args:
        test_case (GetBNGResolutionLabelTestCase): Test case from JSON file.
    """
    # Load test case data
    bng_ref_string = test_case["bng_ref_string"]
    expected = test_case["expected"]
    # Assert that the function returns the expected result
    assert _get_bng_resolution_label(bng_ref_string) == expected


class FormatBNGRefStringTestCase(TypedDict):
    """TypedDict for _format_bng_ref_string function test cases.

    Attributes:
        bng_ref_string (str): The BNG reference string.
        expected (str): The pretty formatted BNG reference string.
    """

    bng_ref_string: str
    expected: str


# Parameterised test for _format_bng_ref_string function
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "bng_reference_test_cases.json"
    )["_format_bng_ref_string"],
)
def test__format_bng_ref_string(test_case: FormatBNGRefStringTestCase) -> None:
    """Test _format_bng_ref_string function with test cases from JSON file.

    Args:
        test_case (FormatBNGRefStringTestCase): Test case from JSON file.
    """
    # Load test case data
    bng_ref_string = test_case["bng_ref_string"]
    expected = test_case["expected"]
    # Assert that the function returns the expected result
    assert _format_bng_ref_string(bng_ref_string) == expected


class BNGReferenceTestCase(TypedDict):
    """TypedDict for _format_bng_ref_string function test cases.

    Attributes:
        bng_ref_string (str): The BNG reference string.
        expected_exception (dict[str, str]): The expected exception is a dictionary with
            the exception name.
        expected_bng_ref_compact (str): The BNG reference with whitespace removed.
        expected_bng_ref_formatted (str): The pretty-formatted version of the BNG
            reference with single spaces between components.
        expected_resolution_metres (int): The resolution of the BNG reference in meters.
        expected_resolution_label (str): The resolution of the BNG reference expressed
            as a descriptive string.
    """

    bng_ref_string: str
    expected_exception: dict[str, str] | None
    expected_bng_ref_compact: str | None
    expected_bng_ref_formatted: str | None
    expected_resolution_metres: int | None
    expected_resolution_label: str | None


# Parameterised test for BNGReference object
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "bng_reference_test_cases.json"
    )["BNGReference"],
)
def test_bngreference(test_case: BNGReferenceTestCase) -> None:
    """Test BNGReference object with test cases from JSON file.

    Args:
        test_case (BNGReferenceTestCase): Test case from JSON file with the following
        keys:
            - bng_ref_string
            - expected_bng_ref_compact
            - expected_bng_ref_formatted
            - expected_resolution_metres
            - expected_resolution_label
    """
    # Load test case data
    bng_ref_string = test_case["bng_ref_string"]

    if "expected_exception" in test_case:
        # Get exception name from test case
        exception_name = test_case["expected_exception"]["name"]
        # Get exception class from name
        exception_class = _EXCEPTION_MAP[exception_name]
        # Assert that the test case raises the expected exception
        with pytest.raises(exception_class):
            BNGReference(bng_ref_string)

    else:
        # Initialise BNGReference object with the test case input
        bng_ref = BNGReference(bng_ref_string)

        # Assert that each property returns the expected value
        assert bng_ref.bng_ref_compact == test_case["expected_bng_ref_compact"]
        assert bng_ref.bng_ref_formatted == test_case["expected_bng_ref_formatted"]
        assert bng_ref.resolution_metres == test_case["expected_resolution_metres"]
        assert bng_ref.resolution_label == test_case["expected_resolution_label"]



"""Configuration file for Sphinx documentation."""

import sys
from datetime import datetime
from pathlib import Path

# Project root directory from docs/
# Represents parent directory of the osbng package
PROJECT_ROOT = Path(__file__).resolve().parent.parent

# Add root directory to sys.path
# Ensures Python can import osbng during autodoc
sys.path.insert(0, str(PROJECT_ROOT))

# Project information
project = "osbng"
author = "Ordnance Survey"
copyright = f"{datetime.now().year}, {author}"
release = "0.5.1"

# Sphinx extensions
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    # Jupyter Notebook support
    "nbsphinx",
    # Copy button for code blocks
    "sphinx_copybutton",
]

# Napoleon configuration
napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_use_admonition_for_examples = True
napoleon_use_admonition_for_notes = True
napoleon_use_admonition_for_references = True

# Only insert the __init__ docstring for class documentation
autoclass_content = "init"

# Templates path
templates_path = ["_templates"]
# Patterns to exclude
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# HTML output configuration
html_theme = "pydata_sphinx_theme"
html_title = "osbng"
html_static_path = ["_static"]
html_theme_options = {
    "logo": {
        "text": "osbng",
        "image_light": "_static/images/os_logo_mono_dark_rgb.svg",
        "image_dark": "_static/images/os_logo_mono_light_rgb.svg",
    },
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/OrdnanceSurvey/osbng-py",
            "icon": "fab fa-github",
            "type": "fontawesome",
        },
    ],
    # Ensure the icon area shows on the right
    "navbar_end": ["theme-switcher", "navbar-icon-links"],
}

# Support cross‑references to objects in external projects
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", "inventories/python.inv"),
    "numpy": ("https://numpy.org/doc/stable/", "inventories/numpy.inv"),
    "shapely": ("https://shapely.readthedocs.io/en/stable/", "inventories/shapely.inv"),
    "geopandas": ("https://geopandas.org/en/stable/", "inventories/geopandas.inv"),
}



"""Provides functionality for traversing the British National Grid (BNG) index system.

It supports spatial analyses such as distance-constrained nearest neighbour searches and
'distance within' queries by offering:

- **Grid traversal**: Generate k-discs and k-rings around a given grid square.
- **Neighbourhood operations**: Identify neighbouring grid squares and checking
  adjacency.
- **Distance computation**: Calculate the distance between grid square centroids.
- **Proximity queries**: Retrieve all grid squares within a specified absolute
  distance.

"""

import warnings

import numpy as np

from osbng.bng_reference import BNGReference, _validate_bngreferences
from osbng.errors import BNGExtentError, BNGNeighbourError
from osbng.hierarchy import bng_to_parent
from osbng.indexing import bng_to_xy, xy_to_bng

__all__ = [
    "bng_kring",
    "bng_kdisc",
    "bng_distance",
    "bng_neighbours",
    "bng_is_neighbour",
    "bng_dwithin",
]


def _ring_or_disc(
    bng_ref: BNGReference, k: int, is_disc: bool, return_relations: bool
) -> list[BNGReference] | list[(BNGReference, int, int)]:
    """Helper function to extract grid squares in a disc or ring.

    Args:
        bng_ref (BNGReference): A BNGReference object.
        k (int): Grid distance in units of grid squares.
        is_disc (bool): If True, returns all grid squares within distance k.  If False,
            only returns the outer ring.
        return_relations (bool): If True, returns a list of (BNGReference, dx, dy)
            tuples where dx, dy are integer offsets in grid units.  If False, returns a
            list of BNGReference objects.

    Returns:
        if return_relations==True:
            list[(BNGReference, dx, dy)]: All BNGReference objects representing grid
                squares in a square ring or disc of radius k, with the x- and y-offsets
                (in grid square units) between bng_ref and each returned BNGReference.
        else:
            list[BNGReference]: All BNGReference objects representing grid squares in a
                square ring or disc of radius k.
    """
    # Check that k is a positive integer
    if k <= 0:
        raise ValueError("k must be a positive integer.")

    # Derive point location of root square
    xc, yc = bng_to_xy(bng_ref, position="centre")

    # Initialise list of ring BNG reference objects
    kring_refs = []

    # Track whether we need to raise an extent warning
    raise_extent_warning = False

    # Iterate over all dx/dy within range
    for dy in range(-k, k + 1)[::-1]:
        for dx in range(-k, k + 1):
            # Include all dx/dy combinations for disks
            # Only include edges for rings
            if is_disc | (abs(dy) == k) | (abs(dx) == k):
                try:
                    ring_ref = xy_to_bng(
                        xc + (dx * bng_ref.resolution_metres),
                        yc + (dy * bng_ref.resolution_metres),
                        bng_ref.resolution_metres,
                    )
                # Catch extent errors and track whether warning is needed
                except BNGExtentError:
                    raise_extent_warning = True
                else:
                    kring_refs.append(
                        (ring_ref, dx, dy)
                    ) if return_relations else kring_refs.append(ring_ref)

    # Raise an extent warning if an error has been caught
    # Note: do this after the above, otherwise repeated warnings will be raised!
    if raise_extent_warning:
        warnings.warn(
            "One or more of the requested grid squares falls outside of the BNG index "
            + "system extent and will not be returned."
        )

    return kring_refs


@_validate_bngreferences
def bng_kring(
    bng_ref: BNGReference, k: int, *, return_relations: bool = False
) -> list[BNGReference] | list[tuple[BNGReference, int, int]]:
    """Returns a hollow ring of BNGReference objects around a ``BNGReference`` object.

    Returns all :class:`~osbng.bng_reference.BNGReference` objects at a grid distance
    ``k``.

    Notes:
        Returned :class:`~osbng.bng_reference.BNGReference` objects are ordered North to
        South then West to East, therefore not in ring order.

    Args:
        bng_ref (BNGReference): A :class:`~osbng.bng_reference.BNGReference` object.
        k (int): Grid distance in units of grid squares.

    Keyword Args:
        return_relations (bool, optional): If True, returns a list of
            (:class:`~osbng.bng_reference.BNGReference`, dx, dy) tuples where dx, dy are
            integer offsets in grid units.  If False (default), returns a list of
            :class:`~osbng.bng_reference.BNGReference` objects.  Keyword only.

    Returns:
        list[BNGReference]: If ``return_relations`` is False (default), returns all
        :class:`~osbng.bng_reference.BNGReference` objects representing squares in a
        square ring of radius ``k``.
        If ``return_relations`` is True, returns a list of
        (:class:`~osbng.bng_reference.BNGReference`, dx, dy) tuples, where dx and dy are
        the x and y offsets between ``bng_ref`` and each returned
        :class:`~osbng.bng_reference.BNGReference` object in units of grid squares.

    Examples:
        >>> bng_kring(BNGReference("SU1234"), 1)
        [
        BNGReference(bng_ref_formatted=SU 11 35, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 12 35, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 13 35, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 11 34, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 13 34, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 11 33, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 12 33, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 13 33, resolution_label=1km)
        ]
        >>> bng_kring(BNGReference("SU1234"), 1, return_relations=True)
        [(BNGReference(bng_ref_formatted=SU 11 35, resolution_label=1km), -1, 1),
        (BNGReference(bng_ref_formatted=SU 12 35, resolution_label=1km), 0, 1),
        (BNGReference(bng_ref_formatted=SU 13 35, resolution_label=1km), 1, 1),
        (BNGReference(bng_ref_formatted=SU 11 34, resolution_label=1km), -1, 0),
        (BNGReference(bng_ref_formatted=SU 13 34, resolution_label=1km), 1, 0),
        (BNGReference(bng_ref_formatted=SU 11 33, resolution_label=1km), -1, -1),
        (BNGReference(bng_ref_formatted=SU 12 33, resolution_label=1km), 0, -1),
        (BNGReference(bng_ref_formatted=SU 13 33, resolution_label=1km), 1, -1)]
        >>> bng_kring(BNGReference("SU1234"), 3)
        [list of 24 BNGReference objects]

    See Also:
        - The equivalent :meth:`osbng.bng_reference.BNGReference.bng_kring` instance
          method.
        - The :func:`~osbng.traversal.bng_kdisc` function and
          :meth:`osbng.bng_reference.BNGReference.bng_kdisc` instance method for
          finding all :class:`~osbng.bng_reference.BNGReference` objects within a
          distance d from an input :class:`~osbng.bng_reference.BNGReference` object.
    """
    return _ring_or_disc(bng_ref, k, False, return_relations)


@_validate_bngreferences
def bng_kdisc(
    bng_ref: BNGReference, k: int, *, return_relations: bool = False
) -> list[BNGReference] | list[tuple[BNGReference, int, int]]:
    """Returns a filled disc of BNGReference objects around a ``BNGReference``.

    Returns all :class:`~osbng.bng_reference.BNGReference` objects up to a grid
    distance ``k``, including the given central
    :class:`~osbng.bng_reference.BNGReference` object.

    Notes:
        Returned :class:`~osbng.bng_reference.BNGReference` objects are ordered North to
        South then West to East.

    Args:
        bng_ref (BNGReference): A :class:`~osbng.bng_reference.BNGReference` object.
        k (int): Grid distance in units of grid squares.

    Keyword Args:
        return_relations (bool, optional): If True, returns a list of
            (:class:`~osbng.bng_reference.BNGReference`, dx, dy) tuples where dx, dy are
            integer offsets in grid units.  If False (default), returns a list of
            :class:`~osbng.bng_reference.BNGReference` objects. Keyword only.

    Returns:
        list[BNGReference]: If ``return_relations`` is False (default), returns all
        :class:`~osbng.bng_reference.BNGReference` objects representing grid squares in
        a square ring of radius k.
        If ``return_relations`` is True, returns a list of
        (:class:`~osbng.bng_reference.BNGReference`, dx, dy) tuples, where dx and dy are
        the x and y offsets between ``bng_ref`` and each returned
        :class:`~osbng.bng_reference.BNGReference` object in units of grid squares.

    Examples:
        >>> bng_kdisc(BNGReference("SU1234"), 1)
        [
        BNGReference(bng_ref_formatted=SU 11 35, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 12 35, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 13 35, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 11 34, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 12 34, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 13 34, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 11 33, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 12 33, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 13 33, resolution_label=1km)
        ]
        >>> bng_kdisc(BNGReference("SU1234"), 1, return_relations=True)
        [(BNGReference(bng_ref_formatted=SU 11 35, resolution_label=1km), -1, 1),
        (BNGReference(bng_ref_formatted=SU 12 35, resolution_label=1km), 0, 1),
        (BNGReference(bng_ref_formatted=SU 13 35, resolution_label=1km), 1, 1),
        (BNGReference(bng_ref_formatted=SU 11 34, resolution_label=1km), -1, 0),
        (BNGReference(bng_ref_formatted=SU 12 34, resolution_label=1km), 0, 0),
        (BNGReference(bng_ref_formatted=SU 13 34, resolution_label=1km), 1, 0),
        (BNGReference(bng_ref_formatted=SU 11 33, resolution_label=1km), -1, -1),
        (BNGReference(bng_ref_formatted=SU 12 33, resolution_label=1km), 0, -1),
        (BNGReference(bng_ref_formatted=SU 13 33, resolution_label=1km), 1, -1)]
        >>> bng_kdisc(BNGReference("SU1234"), 3)
        [list of 49 BNGReference objects]

    See Also:
        - The equivalent :meth:`osbng.bng_reference.BNGReference.bng_kdisc` instance
          method.
        - The :func:`~osbng.traversal.bng_kring` function and
          :meth:`osbng.bng_reference.BNGReference.bng_kring` instance method for
          finding all :class:`~osbng.bng_reference.BNGReference` objects at a
          specific distance d from an input
          :class:`~osbng.bng_reference.BNGReference` object.
    """
    return _ring_or_disc(bng_ref, k, True, return_relations)


@_validate_bngreferences
def bng_distance(
    bng_ref1: BNGReference, bng_ref2: BNGReference, *, edge_to_edge: bool = False
) -> float:
    """Returns the euclidean distance between two ``BNGReference`` objects.

    When ``edge_to_edge`` is False, the distance is the centroid-to-centroid distance in
    metres.  When ``edge_to_edge`` is True, the distance is the shortest distance
    between any two parts of the grid squares.

    Notes:
        Note that the two :class:`~osbng.bng_reference.BNGReference` objects do not
        necessarily need to share a common resolution.  When ``edge_to_edge`` = True and
        ``bng_ref1`` and ``bng_ref2`` have a parent-child relationship, the returned
        distance is 0.

    Args:
        bng_ref1 (BNGReference): A :class:`~osbng.bng_reference.BNGReference` object.
        bng_ref2 (BNGReference): A :class:`~osbng.bng_reference.BNGReference` object.

    Keyword Args:
        edge_to_edge (bool, optional): If False (default), distance will be
            centroid-to-centroid distance.  If True, distance will be the shortest
            distance between any point in the grid squares.  Keyword only.

    Returns:
        float: The euclidean distance between the two
        :class:`~osbng.bng_reference.BNGReference` objects.

    Raises:
        TypeError: If the first or second argument is not a
            :class:`~osbng.bng_reference.BNGReference` object.

    Examples:
        >>> bng_distance(BNGReference("SE1433"), BNGReference("SE1533"))
        1000.0
        >>> bng_distance(
        ...     BNGReference("SE1433"), BNGReference("SE1533"), edge_to_edge=True
        ... )
        0.0
        >>> bng_distance(BNGReference("SE1433"), BNGReference("SE1631"))
        2828.42712474619
        >>> bng_distance(BNGReference("SE1433"), BNGReference("SE"))
        39147.158262126766
        >>> bng_distance(BNGReference("SE1433"), BNGReference("SENW"))
        42807.709586007986
        >>> bng_distance(BNGReference("SE"), BNGReference("OV"))
        141421.35623730952
        >>> bng_distance(BNGReference("SU"), BNGReference("SU2345"), edge_to_edge=True)
        0.0

    See Also:
        - The equivalent :meth:`osbng.bng_reference.BNGReference.bng_distance` instance
          method.
        - The :func:`~osbng.traversal.bng_dwithin` function and
          :meth:`osbng.bng_reference.BNGReference.bng_dwithin` instance method for
          obtaining all :class:`~osbng.bng_reference.BNGReference` objects within a
          defined distance of an input :class:`~osbng.bng_reference.BNGReference`
          object.
    """
    # Catch the special case of parent-child relationship when using edge-to-edge
    if (bng_ref1.resolution_metres != bng_ref2.resolution_metres) & edge_to_edge:
        # Identify the possible parent and child
        # Note this is required, to avoid errors by testing bng_to_parent on a
        # BNGReference with resolution of 100km!
        parent_candidate, child_candidate = (
            (bng_ref1, bng_ref2)
            if bng_ref1.resolution_metres > bng_ref2.resolution_metres
            else (bng_ref2, bng_ref1)
        )

        # Return distance of 0 if one is a parent of the other
        if (
            bng_to_parent(
                child_candidate, resolution=parent_candidate.resolution_metres
            )
            == parent_candidate
        ):
            return 0.0

    # Derive the centroid of the first BNGReference object
    centroid1 = bng_to_xy(bng_ref1, position="centre")

    # Derive the centroid of the second BNGReference object
    centroid2 = bng_to_xy(bng_ref2, position="centre")

    # Note this must be a new if-else logic to the above special case, to catch cases
    # where bng_ref1 and bng_ref2 do not share a resolution but are not parents
    if edge_to_edge:
        # For edge-to-edge distances, the x-distance and y-distance are the
        # centroid-to-centroid distance minus half the box width/height at either end
        dx = (
            0
            if centroid1[0] == centroid2[0]
            else abs(centroid1[0] - centroid2[0])
            - 0.5 * (bng_ref1.resolution_metres + bng_ref2.resolution_metres)
        )
        dy = (
            0
            if centroid1[1] == centroid2[1]
            else abs(centroid1[1] - centroid2[1])
            - 0.5 * (bng_ref1.resolution_metres + bng_ref2.resolution_metres)
        )

    else:
        dx = centroid1[0] - centroid2[0]
        dy = centroid1[1] - centroid2[1]

    return float(np.sqrt(dx**2 + dy**2))


@_validate_bngreferences
def bng_neighbours(bng_ref: BNGReference) -> list[BNGReference]:
    """Returns the four ``BNGReference`` object neighbours of a ``BNGReference``.

    The neighbours are defined as the grid squares immediately North, East, South and
    West of the input grid square sharing an edge with the input
    :class:`~osbng.bng_reference.BNGReference` object.

    Args:
        bng_ref (BNGReference): A :class:`~osbng.bng_reference.BNGReference` object.

    Returns:
        list[BNGReference]: The grid squares immediately North, East, South and West
        of bng_ref, in that order.

    Examples:
        >>> bng_neighbours(BNGReference("SU1234"))
        [
        BNGReference(bng_ref_formatted=SU 12 35, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 13 34, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 12 33, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 11 34, resolution_label=1km)
        ]

    See Also:
        - The equivalent :meth:`osbng.bng_reference.BNGReference.bng_neighbours`
          instance method.
        - The :func:`~osbng.traversal.bng_is_neighbour` function and
          :meth:`osbng.bng_reference.BNGReference.bng_is_neighbour` instance method for
          testing whether two :class:`~osbng.bng_reference.BNGReference` objects are
          neighbours.
    """
    # Get the centroid of the bng square
    x, y = bng_to_xy(bng_ref, position="centre")

    # Initialise a neighbours list
    neighbours_list = []

    # Track whether we need to raise an extent warning
    raise_extent_warning = False

    # Iterate through N,E,S,W neighbours
    for dx, dy in [[0, 1], [1, 0], [0, -1], [-1, 0]]:
        try:
            neighbour = xy_to_bng(
                x + (dx * bng_ref.resolution_metres),
                y + (dy * bng_ref.resolution_metres),
                bng_ref.resolution_metres,
            )
        # Catch extent errors and track whether we need to warn
        except BNGExtentError:
            raise_extent_warning = True
        else:
            neighbours_list.append(neighbour)

    # Raise an extent warning if an error has been caught
    # Note: do this after the above, otherwise repeated warnings will be raised!
    if raise_extent_warning:
        warnings.warn(
            "One or more of the requested grid squares falls outside of the BNG index "
            + "system extent and will not be returned."
        )

    return neighbours_list


@_validate_bngreferences
def bng_is_neighbour(bng_ref1: BNGReference, bng_ref2: BNGReference) -> bool:
    """Tests whether two ``BNGReference`` objects are neighbours.

    Returns True if the two :class:`~osbng.bng_reference.BNGReference` objects are
    neighbours, otherwise False.  Neighbours are defined as grid squares that share an
    edge with the first :class:`~osbng.bng_reference.BNGReference` object.

    Args:
        bng_ref1 (BNGReference): A :class:`~osbng.bng_reference.BNGReference` object.
        bng_ref2 (BNGReference): A :class:`~osbng.bng_reference.BNGReference` object.

    Returns:
        bool: True if the two :class:`~osbng.bng_reference.BNGReference` objects are
        neighbours, otherwise False.

    Raises:
        TypeError: If the first or second argument is not a
            :class:`~osbng.bng_reference.BNGReference` object.
        BNGNeighbourError: If the two :class:`~osbng.bng_reference.BNGReference`
            objects are not at the same resolution.

    Examples:
        >>> bng_is_neighbour(BNGReference("SE1921"), BNGReference("SE1821"))
        True
        >>> bng_is_neighbour(BNGReference("SE1922"), BNGReference("SE1821"))
        False
        >>> bng_is_neighbour(BNGReference("SU1234"), BNGReference("SU1234"))
        False

    See Also:
        - The equivalent :meth:`osbng.bng_reference.BNGReference.bng_is_neighbour`
          instance method.
        - The :func:`~osbng.traversal.bng_neighbours` function and
          :meth:`osbng.bng_reference.BNGReference.bng_neighbours` instance method for
          obtaining all neighbouring :class:`~osbng.bng_reference.BNGReference` objects
          of an input :class:`~osbng.bng_reference.BNGReference` object.

    """
    # Check if the two BNGReference objects are at the same resolution
    if bng_ref1.resolution_metres != bng_ref2.resolution_metres:
        raise BNGNeighbourError(
            "The input BNGReference objects are not the same grid resolution."
            "The inputBNGReference objects must be the same grid resolution."
        )
    # Otherwise check if the two BNGReference objects are neighbours
    else:
        return bng_ref2 in bng_neighbours(bng_ref1)


@_validate_bngreferences
def bng_dwithin(bng_ref: BNGReference, d: int | float) -> list[BNGReference]:
    """Returns a list of ``BNGReference`` objects within an absolute distance ``d``.

    All squares will be returned for which any part of its boundary is within distance
    ``d`` of any part of the input :class:`~osbng.bng_reference.BNGReference`'s
    boundary.

    Args:
        bng_ref (BNGReference): A :class:`~osbng.bng_reference.BNGReference` object.
        d (int or float): The absolute distance d in metres.

    Returns:
        list[BNGReference]: All grid squares which have any part of their geometry
        within distance ``d`` of ``bng_ref``'s geometry

    Examples:
        >>> bng_dwithin(BNGReference("SU1234"), 1000)
        [
        BNGReference(bng_ref_formatted=SU 11 33, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 12 33, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 13 33, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 11 34, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 12 34, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 13 34, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 11 35, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 12 35, resolution_label=1km),
        BNGReference(bng_ref_formatted=SU 13 35, resolution_label=1km)
        ]
        >>> bng_dwithin(BNGReference("SU1234"), 1001)
        [list of 21 BNGReference objects]

    See Also:
        - The equivalent :meth:`osbng.bng_reference.BNGReference.bng_dwithin` instance
          method.
        - The :func:`~osbng.traversal.bng_distance` function and
          :meth:`osbng.bng_reference.BNGReference.bng_distance` instance method for
          calculating the distance between two input
          :class:`~osbng.bng_reference.BNGReference` objects.
    """
    # Convert distance to units of k
    k = int(np.ceil(d / bng_ref.resolution_metres))

    # Get full kdisc
    disc_refs = bng_kdisc(bng_ref, k)

    # Return only those whose centroids are within distance
    return [r for r in disc_refs if bng_distance(bng_ref, r, edge_to_edge=True) <= d]



"""Provides functionality to manipulate British National Grid (BNG) references.

----------------------
BNGReference Object
----------------------

The BNG index system uses BNG references, also known more simply as grid or tile
references, to identify and index locations across Great Britain (GB) into grid squares
at various resolutions.

The :class:`~osbng.bng_reference.BNGReference` object is a custom class that
encapsulates a BNG reference string, providing properties and methods to access
and manipulate the reference.

---------------------------------------
British National Grid Index System
---------------------------------------

The Ordnance Survey (OS) BNG index system, also known as the OS National Grid, is a
rectangular Cartesian 700 x 1300km grid system based upon the transverse Mercator
projection. In the BNG, locations are specified using coordinates, ``eastings (x)`` and
``northings (y)``, measured in meters from a defined origin point (0, 0) southwest of
the Isles of Scilly off the coast of Cornwall, England. Values increase to the
northeast, covering all of mainland GB and surrounding islands.

The BNG is structured using a hierarchical system of grid squares at various
resolutions. At its highest level, the grid divides GB into 100km by 100km squares,
each identified by a two-letter code. Successive levels of resolution further subdivide
the grid squares into finer detail, down to individual 1-meter squares.

.. image:: https://raw.githubusercontent.com/OrdnanceSurvey/osbng-py/main/docs/_static/images/osbng_grids_100km.png
   :align: center

---------------------------
BNG Reference Structure
---------------------------

Each BNG reference string consists of a series of alphanumeric characters that encode
the easting and northing at a given resolution.

A BNG reference includes a 2-letter prefix that identifies the 100km grid square. This
is followed by an easting and northing value, and optionally, a suffix indicating an
ordinal (intercardinal) direction (``NE``, ``SE``, ``SW``, ``NW``). These suffixes
represent a quadtree subdivision of the grid at the 'standard' resolutions (``100km``,
``10km``, ``1km``, ``100m``, and ``10m``), with each direction indicating a specific
quadrant.::

<prefix><easting value><northing value><suffix>

There are two exceptions to this structure:

1.  At the 100km resolution, a BNG reference consists only of the prefix.
2.  At the 50km resolution, a BNG reference includes the prefix and the ordinal
    direction suffix but does not include easting or northing components.

A BNG reference can be expressed at different scales, as follows:

=========== ========================================================= ==================
Resolution   Description                                               Example
=========== ========================================================= ==================
100km        Identified by a two-letter code                           TQ
50km         Subdivides the 100km grid into four quadrants. The grid   TQ SW
             reference adds an ordinal direction suffix
             (NE, NW, SE, SW) to indicate the quadrant within the
             100km square.
10km         Adds one-digit easting and northing values                TQ 2 3
5km          Subdivides the 10km square adding an ordinal suffix       TQ 23 SW
1km          Adds two-digit easting and northing values                TQ 23 34
500m         Subdivides the 1km square adding an ordinal suffix        TQ 23 34 NE
100m         Adds three-digit easting and northing values              TQ 238 347
50m          Subdivides the 100m square adding an ordinal suffix       TQ 238 347 SE
10m          Adds four-digit easting and northing values               TQ 2386 3472
5m           Subdivides the 10m square adding an ordinal suffix        TQ 2386 3472 NW
1m           Adds five-digit easting and northing values               TQ 23863 34729
=========== ========================================================= ==================

---------------------------
BNG Reference Formatting
---------------------------

BNG reference strings passed to a :class:`~osbng.bng_reference.BNGReference` object must
adhere to the following format:

- Whitespace may or may not separate the components of the reference (i.e. between the
  two-letter 100km grid square prefix, easting, northing, and ordinal suffix).
- If whitespace is present, it should be a single space character.
- Whitespace can be inconsistently used between components of the reference.
- The two-letter 100km grid square prefixes and ordinal direction suffixes
  (``NE``, ``SE``, ``SW``, ``NW``) should be capitalised.

-----------------------------------------------
  EPSG:27700 (OSGB36 / British National Grid)
-----------------------------------------------

The BNG system is a practical application of the `EPSG:27700 (OSGB36 / British National
Grid) <https://epsg.io/27700>`__ coordinate reference system which provides the geodetic
framework that defines how locations defined by easting and northing coordinates and
encoded as BNG references (e.g. 'ST 569 714') are projected to the grid.

----------------------------
BNG Reference Application
----------------------------

The BNG index system is widely used by the geospatial community across GB. At each
resolution, a given location can be identified with increasing detail, allowing for
variable accuracy depending on the geospatial application, from small-scale mapping to
precise survey measurements.
"""

import inspect
import re
from functools import wraps
from typing import Callable, Union

from shapely.geometry import Polygon, mapping

from osbng.errors import BNGReferenceError
from osbng.resolution import BNG_RESOLUTIONS

__all__ = ["BNGReference"]

# Compile regular expression pattern for BNG reference string validation
# The geographical extent of the BNG reference system is defined as:
# 0 <= easting < 700000 and 0 <= northing < 1300000
# Supports the following resolutions:
# 100km, 50km, 10km, 5km, 1km, 500m, 100m, 50m, 10m, 5m, 1m
_PATTERN = re.compile(
    r"""
    ^
    # 100km grid square prefix
    (H[LMNOPQRSTUVWXYZ]|
     N[ABCDEFGHJKLMNOPQRSTUVWXYZ]|
     O[ABFGLMQRVW]|
     S[ABCDEFGHJKLMNOPQRSTUVWXYZ]|
     T[ABFGLMQRVW]|
     J[LMQRVW])
    # Zero or one whitespace characters
    \s?
    # Easting and northing coordinates
    # 2-8 digit BNG reference
    # Not separated by whitespace
    (?:(\d{2}|\d{4}|\d{6}|\d{8}|
    # Separated by whitespace
    \d{1}\s\d{1}|\d{2}\s\d{2}|\d{3}\s\d{3}|\d{4}\s\d{4}|
    # 10-digit BNG reference
    # Not separated by whitespace
    \d{10}$|
    # Separated by whitespace
    \d{5}\s\d{5}$))?
    # Zero or one whitespace characters
    \s?
    # Ordinal direction suffix
    (NE|SE|SW|NW)?$""",
    re.VERBOSE,
)


def _validate_bng_ref_string(bng_ref_string: str) -> bool:
    """Validates a BNG reference string using a regular expression pattern.

    Args:
        bng_ref_string (str): The BNG reference string to validate.

    Returns:
        bool: True if the BNG reference is valid, False otherwise.

    Examples:
        >>> _validate_bng("TQ 12 34")
        True
        >>> _validate_bng("TQ1234")
        True
        >>> _validate_bng("tq123")
        False
    """
    return bool(_PATTERN.match(bng_ref_string))


def _get_bng_resolution_metres(bng_ref_string: str) -> int:
    """Returns the resolution of a BNG reference string in metres.

    Args:
        bng_ref_string (str): The BNG reference string.

    Returns:
        resolution (int): The resolution of the BNG reference in metres.

    Examples:
        >>> _get_bng_resolution_metres("TQ1234")
        1000
        >>> _get_bng_resolution_metres("TQ12")
        10000
        >>> _get_bng_resolution_metres("TQSW")
        50000
    """
    # Match BNG reference string against regex pattern
    match = _PATTERN.match(bng_ref_string)

    # Extract components of the BNG reference
    en_components = match.group(2)
    suffix = match.group(3)

    # Determine resolution based on length of easting and northing components
    # and whether an ordinal suffix is present.
    if en_components is None:
        resolution = 100000
    else:
        length = len(en_components)
        # The possible resolutions are powers of ten: 1, 10, 100, 1000, 10000, 100000
        # Integer division by 2 to determine the appropriate power of ten
        # Subtracting from 5 aligns the length with the correct power of ten
        resolution = 10 ** (5 - length // 2)

    # Adjust for ordinal suffix if present
    if suffix:
        resolution //= 2  # Ordinal suffix halves the resolution

    return resolution


def _get_bng_resolution_label(bng_ref_string: str) -> str:
    """Returns the resolution of a BNG reference string as a descriptive label.

    The resolution is returned in a human-readable format, such as '10km', '50km', '5km'
    etc.

    Args:
        bng_ref_string (str): The BNG reference string.

    Returns:
        str: The resolution of the BNG reference as a string label.

    Examples:
        >>> _get_bng_resolution_label("TQ1234")
        '1km'
        >>> _get_bng_resolution_label("TQ12")
        '10km'
        >>> _get_bng_resolution_label("TQSW")
        '50km'
    """
    # Get the resolution in meters
    resolution_meters = _get_bng_resolution_metres(bng_ref_string)

    # Get the resolution label
    return BNG_RESOLUTIONS.get(resolution_meters)["label"]


def _format_bng_ref_string(bng_ref_string: str) -> str:
    """Returns a BNG reference string in pretty format.

    Uses a single space between the prefix, easting, northing, and suffix to improve
    readability.

    Args:
        bng_ref_string (str): The BNG reference string.

    Returns:
        pretty_format (str): The pretty formatted BNG reference string.

    Examples:
        >>> _format_bng_ref_string("TQ1234")
        'TQ 12 34'
        >>> _format_bng_ref_string("TQ1234NE")
        'TQ 12 34 NE'
        >>> _format_bng_ref_string("TQ127349NE")
        'TQ 127 349 NE'
    """
    # Match BNG reference string against regex pattern
    match = _PATTERN.match(bng_ref_string)

    # Extract components of the BNG reference string
    prefix = match.group(1)
    en_components = match.group(2)
    suffix = match.group(3)

    # Pretty format the BNG reference string
    if en_components is None:
        pretty_format = prefix
    else:
        # Split easting and northing components
        length = len(en_components)
        easting = en_components[: length // 2]
        northing = en_components[length // 2 :]
        # Add whitespace between components
        pretty_format = f"{prefix} {easting} {northing}"

    # Add ordinal suffix if present
    if suffix:
        pretty_format += f" {suffix}"

    return pretty_format


class BNGReference:
    """A custom object for handling BNG references.

    Converts a BNG reference string into a :class:`~osbng.bng_reference.BNGReference`
    object, ensuring type consistency across the package. All functions accepting or
    returning BNG references enforce the use of this class.

    These functions are available both as instance methods of the
    :class:`~osbng.bng_reference.BNGReference` object and as standalone functions,
    providing users with the flexibility to either:

    - Create a :class:`~osbng.bng_reference.BNGReference` object and pass it to a
      function.
    - Create a :class:`~osbng.bng_reference.BNGReference` object and use one of its
      instance methods.

    Args:
        bng_ref_string (str): The BNG reference string.

    Attributes:
        bng_ref_compact (str): The BNG reference string of this ``BNGReference`` with
            whitespace removed.
        bng_ref_formatted (str): The pretty-formatted BNG reference string of this
            ``BNGReference`` with single spaces between components.
        resolution_metres (int): The resolution of this ``BNGReference`` in meters.
        resolution_label (str): The resolution of this ``BNGReference`` expressed as a
            descriptive string.
        __geo_interface__ (dict): A GeoJSON-like mapping of this ``BNGReference``.

    Methods:
        bng_to_xy(position: str = "lower-left") -> tuple[int | float, int | float]:
            Returns easting and northing coordinates of this  ``BNGReference`` at a
            specified grid square position.
        bng_to_bbox() -> tuple[int, int, int, int]: Returns grid square bounding box
            coordinates of this ``BNGReference``.
        bng_to_grid_geom() -> Polygon: Returns a grid square as a ``Shapely Polygon``
            of this ``BNGReference``.
        bng_to_children(resolution: int | str | None = None) -> list[BNGReference]:
            Returns a list of ``BNGReference`` objects that are children of this
            ``BNGReference``.
        bng_to_parent(resolution: int | str | None = None) -> BNGReference: Returns the
            ``BNGReference`` that is the parent of this ``BNGReference``.
        bng_kring(k: int, return_relations: bool = False) -> list[BNGReference] |
            list[tuple[BNGReference, int, int]]:
            Returns a list of ``BNGReference`` objects forming a hollow ring around this
            ``BNGReference``.
        bng_kdisc(k: int, return_relations: bool = False) -> list[BNGReference] |
            list[tuple[BNGReference, int, int]]:
            Returns a list of ``BNGReference`` objects forming a filled disc around this
            ``BNGReference``.
        bng_distance(bng_ref2: BNGReference, edge_to_edge: bool = False) -> float:
            Returns the euclidean distance between ``bng_ref2`` and this
            ``BNGReference``.
        bng_neighbours() -> list[BNGReference]: Returns a list of ``BNGReference``
            objects representing the four neighbouring grid squares sharing an edge
            with this ``BNGReference``.
        bng_is_neighbour(bng_ref2: BNGReference) -> bool: Tests whether ``bng_ref2`` is
            a neighbour of this ``BNGReference``.
        bng_dwithin(d:int | float) -> list[BNGReference]: Returns a list of
            ``BNGReference`` objects within a distance ``d`` from this ``BNGReference``.

    Raises:
        BNGReferenceError: If the BNG reference string is invalid.

    Examples:
        >>> bng_ref = BNGReference("TQ1234")
        >>> bng_ref.bng_ref_compact
        'TQ1234'
        >>> bng_ref.bng_ref_formatted
        'TQ 12 34'
        >>> bng_ref.resolution_metres
        1000
        >>> bng_ref.resolution_label
        '1km'
        >>> bng_ref.bng_to_xy()
        (512000, 134000)
        >>> bng_ref.bng_to_bbox()
        (512000, 134000, 513000, 135000)
        >>> bng_ref.bng_to_parent()
        BNGReference(bng_ref_formatted=TQ 1 3 SW, resolution_label=5km)
        >>> bng_ref.bng_neighbours()
        [BNGReference(bng_ref_formatted=TQ 12 35, resolution_label=1km),
         BNGReference(bng_ref_formatted=TQ 13 34, resolution_label=1km),
         BNGReference(bng_ref_formatted=TQ 12 33, resolution_label=1km),
         BNGReference(bng_ref_formatted=TQ 11 34, resolution_label=1km)]
    """

    def __init__(self, bng_ref_string: str):
        """Initialises a ``BNGReference`` from a BNG reference string."""
        # Validate the BNG reference string
        if not _validate_bng_ref_string(bng_ref_string):
            raise BNGReferenceError(f"Invalid BNG reference string: '{bng_ref_string}'")

        # Remove all whitespace for internal storage
        self._bng_ref_compact = bng_ref_string.replace(" ", "")

    @property
    def bng_ref_compact(self) -> str:
        """The BNG reference string of this ``BNGReference`` with whitespace removed."""
        return self._bng_ref_compact

    @property
    def bng_ref_formatted(self) -> str:
        """The BNG reference string of this ``BNGReference`` in pretty format.

        Uses a single space between the prefix, easting, northing, and suffix to
        improve readability.
        """
        return _format_bng_ref_string(self._bng_ref_compact)

    @property
    def resolution_metres(self) -> int:
        """The resolution of this ``BNGReference`` in meters."""
        return _get_bng_resolution_metres(self._bng_ref_compact)

    @property
    def resolution_label(self) -> str:
        """The resolution of this ``BNGReference`` expressed as a string label.

        The resolution is returned in a human-readable format, such as '10km', '50km',
        '5km', etc.

        See Also:
            :data:`osbng.resolution.BNG_RESOLUTIONS` for mappings from metre-based
            integer resolution values to string label representations.
        """
        return _get_bng_resolution_label(self._bng_ref_compact)

    @property
    def __geo_interface__(self) -> dict[str, Union[str, dict]]:
        """A GeoJSON-like mapping of this ``BNGReference``.

        Implements the `__geo_interface__
        <https://gist.github.com/sgillies/2217756>`__ protocol. The returned data
        structure represents the :class:`~osbng.bng_reference.BNGReference` object as a
        GeoJSON-like Feature.
        """
        return {
            "type": "Feature",
            "properties": {
                "bng_ref": self.bng_ref_compact,
            },
            "geometry": mapping(self.bng_to_grid_geom()),
        }

    def __eq__(self, other: object):
        """Determines whether this ``BNGReference`` is equal to ``other``."""
        if isinstance(other, BNGReference):
            return self.bng_ref_compact == other.bng_ref_compact
        return False

    def __lt__(self, other: object):
        """Determines whether this ``BNGReference`` is ordered before ``other``.

        For two :class:`~osbng.bng_reference.BNGReference` objects, ordering is done in
        the following order:

        1. Rank by :attr:`~osbng.bng_reference.BNGReference.resolution_metres`, where
           higher resolutions are ordered first.
        2. Rank by :attr:`~osbng.bng_reference.BNGReference.bng_ref_compact`
           alphabetically.

        Example:
            >>> BNGReference("SU") < BNGReference("SU1234")
            True
            >>> BNGReference("SU") < BNGReference("TU")
            True
        """
        if isinstance(other, BNGReference):
            return (-self.resolution_metres, self.bng_ref_compact) < (
                -other.resolution_metres,
                other.bng_ref_compact,
            )
        return NotImplemented

    def __hash__(self):
        """Returns a hash value of this ``BNGReference``."""
        return hash(self.bng_ref_compact)

    def __repr__(self):
        """Returns the string representation of this ``BNGReference``."""
        return (
            f"BNGReference(bng_ref_formatted={self.bng_ref_formatted}, "
            f"resolution_label={self.resolution_label})"
        )

    def bng_to_xy(
        self, *, position: str = "lower-left"
    ) -> tuple[int | float, int | float]:
        """Returns easting and northing coordinates of this ``BNGReference``.

        An optional grid square ``position`` keyword argument can be specified to
        return the coordinates of a specific corner or the centre of the grid square.

        Keyword Args:
            position (str, optional): The grid square position expressed as a string.
                One of: 'lower-left', 'upper-left', 'upper-right', 'lower-right',
                'centre'.

        Returns:
            tuple[int | float, int | float]: Easting and northing coordinates as a
                tuple.

        Raises:
            ValueError: If invalid position provided.

        Example:
            >>> BNGReference("SU").bng_to_xy()
            (400000, 100000)
            >>> BNGReference("SU 3 1").bng_to_xy()
            (430000, 110000)
            >>> BNGReference("SU 3 1 NE").bng_to_xy("centre")
            (437500, 117500)
            >>> BNGReference("SU 37289 15541").bng_to_xy("centre")
            (437289.5, 115541.5)

        See Also:
            The equivalent :func:`osbng.indexing.bng_to_xy` function.
        """
        from osbng.indexing import bng_to_xy as _bng_to_xy

        return _bng_to_xy(self, position=position)

    def bng_to_bbox(self) -> tuple[int, int, int, int]:
        """Returns grid square bounding box coordinates of this ``BNGReference``.

        Returns:
            tuple[int, int, int, int]: The grid square bounding box coordinates as a
                tuple.

        Example:
            >>> BNGReference("SU").bng_to_bbox()
            (400000, 100000, 500000, 200000)
            >>> BNGReference("SU 3 1").bng_to_bbox()
            (430000, 110000, 440000, 120000)
            >>> BNGReference("SU 3 1 NE").bng_to_bbox()
            (435000, 115000, 440000, 120000)
            >>> BNGReference("SU 37289 15541").bng_to_bbox()
            (437289, 115541, 437290, 115542)

        See Also:
            The equivalent :func:`osbng.indexing.bng_to_bbox` function.
        """
        from osbng.indexing import bng_to_bbox as _bng_to_bbox

        return _bng_to_bbox(self)

    def bng_to_grid_geom(self) -> Polygon:
        """Returns a grid square as a ``Shapely Polygon`` of this ``BNGReference``.

        Returns:
            Polygon: Grid square as ``Shapely Polygon`` object.

        Example:
            >>> BNGReference("SU").bng_to_grid_geom().wkt
            'POLYGON ((500000 100000, 500000 200000, 400000 200000, 400000 100000,
                500000 100000))'
            >>> BNGReference("SU 3 1").bng_to_grid_geom().wkt
            'POLYGON ((440000 110000, 440000 120000, 430000 120000, 430000 110000,
                440000 110000))'
            >>> BNGReference("SU 3 1 NE").bng_to_grid_geom().wkt
            'POLYGON ((440000 115000, 440000 120000, 435000 120000, 435000 115000,
                440000 115000))'
            >>> BNGReference("SU 37289 15541").bng_to_grid_geom().wkt
            'POLYGON ((437290 115541, 437290 115542, 437289 115542, 437289 115541,
                437290 115541))'
        """
        from osbng.indexing import bng_to_grid_geom as _bng_to_grid_geom

        return _bng_to_grid_geom(self)

    def bng_to_children(
        self, *, resolution: int | str | None = None
    ) -> list["BNGReference"]:
        """Returns a list of child ``BNGReference`` objects of this ``BNGReference``.

        By default, the children of the :class:`~osbng.bng_reference.BNGReference`
        object is defined as the :class:`~osbng.bng_reference.BNGReference` objects in
        the next resolution down from the current ``BNGReference`` resolution. For
        example, 100km -> 50km.

        Notes:
            Any valid resolution can be provided as the child resolution, provided it
            is less than the resolution of the current
            :class:`~osbng.bng_reference.BNGReference` object.

        Keyword Args:
            resolution (int | str | None, optional): The resolution of the children
                :class:`~osbng.bng_reference.BNGReference` objects expressed either
                as a metre-based integer or as a string label. Defaults to None.

        Returns:
            list[BNGReference]: A list of BNGReference objects that are children of the
                current :class:`~osbng.bng_reference.BNGReference` object.

        Raises:
            BNGHierarchyError: If the resolution of the current
                :class:`~osbng.bng_reference.BNGReference` object is 1m.
            BNGHierarchyError: If the resolution is greater than or equal to the
                resolution of the current :class:`~osbng.bng_reference.BNGReference`
                object.
            BNGResolutionError: If an invalid resolution is provided.

        Examples:
            >>> BNGReference("SU").bng_to_children()
            [BNGReference(bng_ref_formatted=SU SW, resolution_label=50km),
            BNGReference(bng_ref_formatted=SU SE, resolution_label=50km),
            BNGReference(bng_ref_formatted=SU NW, resolution_label=50km),
            BNGReference(bng_ref_formatted=SU NE, resolution_label=50km)]
            >>> BNGReference("SU36").bng_to_children()
            [BNGReference(bng_ref_formatted=SU 3 6 SW, resolution_label=5km),
            BNGReference(bng_ref_formatted=SU 3 6 SE, resolution_label=5km),
            BNGReference(bng_ref_formatted=SU 3 6 NW, resolution_label=5km),
            BNGReference(bng_ref_formatted=SU 3 6 NE, resolution_label=5km)]

        See Also:
            The equivalent :func:`osbng.hierarchy.bng_to_children` function.
        """
        from osbng.hierarchy import bng_to_children as _bng_to_children

        return _bng_to_children(self, resolution=resolution)

    def bng_to_parent(self, *, resolution: int | str | None = None) -> "BNGReference":
        """Returns the ``BNGReference`` that is the parent of this ``BNGReference``.

        By default, the parent of the :class:`~osbng.bng_reference.BNGReference` object
        is defined as the :class:`~osbng.bng_reference.BNGReference` in the next BNG
        resolution up from the current :class:`~osbng.bng_reference.BNGReference`
        resolution. For example, 50km -> 100km.

        Notes:
            Any valid resolution can be provided as the parent resolution, provided it
            is greater than the resolution of the current
            :class:`~osbng.bng_reference.BNGReference` object.

        Keyword Args:
            resolution (int | str | None, optional): The resolution of the parent
                :class:`~osbng.bng_reference.BNGReference` objects expressed either as
                a metre-based integer or as a string label. Defaults to None.

        Returns:
            BNGReference: A :class:`~osbng.bng_reference.BNGReference` object that is
                the parent of the current :class:`~osbng.bng_reference.BNGReference`
                object.

        Raises:
            BNGHierarchyError: If the resolution of the current
                :class:`~osbng.bng_reference.BNGReference` object is 100km.
            BNGHierarchyError: If the resolution is less than or equal to the resolution
                of the current :class:`~osbng.bng_reference.BNGReference` object.
            BNGResolutionError: If an invalid resolution is provided.

        Examples:
            >>> BNGReference("SU 3 6 SW").bng_to_parent()
            BNGReference(bng_ref_formatted=SU 3 6, resolution_label=10km)
            >>> BNGReference("SU 342 567").bng_to_parent()
            BNGReference(bng_ref_formatted=SU 34 56 NW, resolution_label=500m)
            >>> BNGReference("SU 342 567").bng_to_parent(resolution=10000)
            BNGReference(bng_ref_formatted=SU 3 5, resolution_label=10km)

        See Also:
            The equivalent :func:`osbng.hierarchy.bng_to_parent` function.
        """
        from osbng.hierarchy import bng_to_parent as _bng_to_parent

        return _bng_to_parent(self, resolution=resolution)

    def bng_kring(
        self, k: int, *, return_relations: bool = False
    ) -> list["BNGReference"]:
        """Returns a hollow ring of BNGReference objects around this ``BNGReference``.

        Returns all :class:`~osbng.bng_reference.BNGReference` objects at a grid
        distance ``k``.

        Notes:
            Returned :class:`~osbng.bng_reference.BNGReference` objects are ordered
            North to South then West to East, therefore not in ring order.

        Args:
            k (int): Grid distance in units of grid squares.

        Keyword Args:
            return_relations (bool, optional): If True, returns a list of
                (BNGReference, dx, dy) tuples where dx, dy are integer offsets in grid
                units.  If False (default), returns a list of
                :class:`~osbng.bng_reference.BNGReference` objects.

        Returns:
            list[BNGReference]: All :class:`~osbng.bng_reference.BNGReference` objects
            representing squares in a square ring of radius k.

        Examples:
            >>> BNGReference("SU1234").bng_kring(1)
            [
            BNGReference(bng_ref_formatted=SU 11 35, resolution_label=1km),
            BNGReference(bng_ref_formatted=SU 12 35, resolution_label=1km),
            BNGReference(bng_ref_formatted=SU 13 35, resolution_label=1km),
            BNGReference(bng_ref_formatted=SU 11 34, resolution_label=1km),
            BNGReference(bng_ref_formatted=SU 13 34, resolution_label=1km),
            BNGReference(bng_ref_formatted=SU 11 33, resolution_label=1km),
            BNGReference(bng_ref_formatted=SU 12 33, resolution_label=1km),
            BNGReference(bng_ref_formatted=SU 13 33, resolution_label=1km)
            ]
            >>> BNGReference("SU1234").bng_kring(1, return_relations=True)
            [
            (BNGReference(bng_ref_formatted=SU 11 35, resolution_label=1km), -1, 1),
            (BNGReference(bng_ref_formatted=SU 12 35, resolution_label=1km), 0, 1),
            (BNGReference(bng_ref_formatted=SU 13 35, resolution_label=1km), 1, 1),
            (BNGReference(bng_ref_formatted=SU 11 34, resolution_label=1km), -1, 0),
            (BNGReference(bng_ref_formatted=SU 13 34, resolution_label=1km), 1, 0),
            (BNGReference(bng_ref_formatted=SU 11 33, resolution_label=1km), -1, -1),
            (BNGReference(bng_ref_formatted=SU 12 33, resolution_label=1km), 0, -1),
            (BNGReference(bng_ref_formatted=SU 13 33, resolution_label=1km), 1, -1)
            ]
            >>> BNGReference("SU1234").bng_kring(3)
            [list of 24 BNGReference objects]

        See Also:
            The equivalent :func:`osbng.traversal.bng_kring` function.
        """
        from osbng.traversal import bng_kring as _bng_kring

        return _bng_kring(self, k, return_relations=return_relations)

    def bng_kdisc(
        self, k: int, *, return_relations: bool = False
    ) -> list["BNGReference"]:
        """Returns a filled disc of BNGReference objects around this ``BNGReference``.

        Returns all :class:`~osbng.bng_reference.BNGReference` objects up to a grid
        distance ``k``, including the given central
        :class:`~osbng.bng_reference.BNGReference` object.

        Notes:
            Returned :class:`~osbng.bng_reference.BNGReference` objects are ordered
            North to South then West to East.

        Args:
            k (int): Grid distance in units of grid squares.

        Keyword Args:
            return_relations (bool, optional): If True, returns a list of
                (BNGReference, dx, dy) tuples where dx, dy are integer offsets in grid
                units.  If False (default), returns a list of
                :class:`~osbng.bng_reference.BNGReference` objects.

        Returns:
            list[BNGReference]: All :class:`~osbng.bng_reference.BNGReference` objects
                representing grid squares in a square of radius ``k``.

        Examples:
            >>> BNGReference("SU1234").bng_kdisc(1)
            [
            BNGReference(bng_ref_formatted=SU 11 35, resolution_label=1km),
            BNGReference(bng_ref_formatted=SU 12 35, resolution_label=1km),
            BNGReference(bng_ref_formatted=SU 13 35, resolution_label=1km),
            BNGReference(bng_ref_formatted=SU 11 34, resolution_label=1km),
            BNGReference(bng_ref_formatted=SU 12 34, resolution_label=1km),
            BNGReference(bng_ref_formatted=SU 13 34, resolution_label=1km),
            BNGReference(bng_ref_formatted=SU 11 33, resolution_label=1km),
            BNGReference(bng_ref_formatted=SU 12 33, resolution_label=1km),
            BNGReference(bng_ref_formatted=SU 13 33, resolution_label=1km)
            ]
            >>> BNGReference("SU1234").bng_kdisc(1, return_relations=True)
            [
            (BNGReference(bng_ref_formatted=SU 11 35, resolution_label=1km), -1, 1),
            (BNGReference(bng_ref_formatted=SU 12 35, resolution_label=1km), 0, 1),
            (BNGReference(bng_ref_formatted=SU 13 35, resolution_label=1km), 1, 1),
            (BNGReference(bng_ref_formatted=SU 11 34, resolution_label=1km), -1, 0),
            (BNGReference(bng_ref_formatted=SU 12 34, resolution_label=1km), 0, 0),
            (BNGReference(bng_ref_formatted=SU 13 34, resolution_label=1km), 1, 0),
            (BNGReference(bng_ref_formatted=SU 11 33, resolution_label=1km), -1, -1),
            (BNGReference(bng_ref_formatted=SU 12 33, resolution_label=1km), 0, -1),
            (BNGReference(bng_ref_formatted=SU 13 33, resolution_label=1km), 1, -1)
            ]
            >>> BNGReference("SU1234").bng_kdisc(3)
            [list of 49 BNGReference objects]

        See Also:
            The equivalent :func:`osbng.traversal.bng_kdisc` function.
        """
        from osbng.traversal import bng_kdisc as _bng_kdisc

        return _bng_kdisc(self, k, return_relations=return_relations)

    def bng_distance(
        self, bng_ref2: "BNGReference", *, edge_to_edge: bool = False
    ) -> float:
        """Returns the euclidean distance between bng_ref2 and this ``BNGReference``.

        When ``edge_to_edge`` is False, the distance is the centroid-to-centroid
        distance in metres.  When ``edge_to_edge`` is True, the distance is the
        shortest distance between any two parts of the grid squares.

        Notes:
            The other :class:`~osbng.bng_reference.BNGReference` object does not
            necessarily need to share a common resolution.  When ``edge_to_edge``
            = True and the two :class:`~osbng.bng_reference.BNGReference` objects have
            a parent-child relationship, the returned distance is 0.

        Args:
            bng_ref2 (BNGReference): A :class:`~osbng.bng_reference.BNGReference` object
                .

        Keyword Args:
            edge_to_edge (bool, optional): If False (default), distance will be
                centroid-to-centroid distance.  If True, distance will be the shortest
                distance between any point in the grid squares.

        Returns:
            float: The euclidean distance between the centroids of the two
            :class:`~osbng.bng_reference.BNGReference` objects.

        Raises:
            TypeError: If the ``bng_ref2`` argument is not a
            :class:`~osbng.bng_reference.BNGReference` object.

        Examples:
            >>> BNGReference("SE1433").bng_distance(BNGReference("SE1533"))
            1000.0
            >>> BNGReference("SE1433").bng_distance(
            ...     BNGReference("SE1533"), edge_to_edge=True
            ... )
            0.0
            >>> BNGReference("SE1433").bng_distance(BNGRerence("SE1631"))
            2828.42712474619
            >>> BNGReference("SE1433").bng_distance(BNGRerence("SE"))
            39147.158262126766
            >>> BNGReference("SE1433").bng_distance(BNGRerence("SENW"))
            42807.709586007986
            >>> BNGReference("SE").bng_distance(BNGRerence("OV"))
            141421.35623730952
            >>> BNGReference("SU").bng_distance(
            ...     BNGReference("SU2345"), edge_to_edge=True
            ... )
            0.0

        See Also:
            The equivalent :func:`osbng.traversal.bng_distance` function.
        """
        from osbng.traversal import bng_distance as _bng_distance

        return _bng_distance(self, bng_ref2, edge_to_edge=edge_to_edge)

    def bng_neighbours(self) -> list["BNGReference"]:
        """Returns the four BNGReference object neighbours to this BNGReference.

        The neighbours are defined as the grid squares immediately North, East, South
        and West of the input grid square sharing an edge with the input
        :class:`~osbng.bng_reference.BNGReference` object.

        Returns:
            list[BNGReference]: The grid squares immediately North, South, East and
            West of this :class:`~osbng.bng_reference.BNGReference` object.

        Examples:
            >>> BNGReference("SU1234").bng_neighbours()
            [BNGReference('SU1235'), BNGReference('SU1334'),
            BNGReference('SU1233'), BNGReference('SU1134')]

        See Also:
            The equivalent :func:`osbng.traversal.bng_neighbours` function.
        """
        from osbng.traversal import bng_neighbours as _bng_neighbours

        return _bng_neighbours(self)

    def bng_is_neighbour(self, bng_ref2: "BNGReference") -> bool:
        """Tests whether ``bng_ref2`` is a neighbour of this ``BNGReference``.

        Neighbours are defined as grid squares that share an edge with this
        :class:`~osbng.bng_reference.BNGReference` object.

        Args:
            bng_ref2 (BNGReference): A :class:`~osbng.bng_reference.BNGReference` object
            .

        Returns:
            bool: True if the two :class:`~osbng.bng_reference.BNGReference` objects are
            neighbours, otherwise False.

        Raises:
            TypeError: If the ``bng_ref2`` argument is not a
            :class:`~osbng.bng_reference.BNGReference` object.
            BNGNeighbourError: If the :class:`~osbng.bng_reference.BNGReference` object
            is not at the same resolution.

        Examples:
            >>> BNGReference("SE1921").bng_is_neighbour(BNGReference("SE1821"))
            True
            >>> BNGReference("SE1922").bng_is_neighbour(BNGReference("SE1821"))
            False
            >>> BNGReference("SU1234").bng_is_neighbour(BNGReference("SU1234"))
            False

        See Also:
            The equivalent :func:`osbng.traversal.bng_is_neighbour` function.
        """
        from osbng.traversal import bng_is_neighbour as _bng_is_neighbour

        return _bng_is_neighbour(self, bng_ref2)

    def bng_dwithin(self, d: int | float) -> list["BNGReference"]:
        """Returns all BNGReference objects within distance ``d`` of this BNGReference.

        All grid squares will be returned for which any part of its boundary is within
        distance ``d`` of any part of the :class:`~osbng.bng_reference.BNGReference`
        object's boundary.

        Args:
            d (int or float): The absolute distance ``d`` in metres.

        Returns:
            list[BNGReference]: All grid squares which have any part of their geometry
            within distance ``d`` of the current grid square

        Examples:
            >>> BNGReference("SU1234").bng_dwithin(1000)
            [
            BNGReference(bng_ref_formatted=SU 11 33, resolution_label=1km),
            BNGReference(bng_ref_formatted=SU 12 33, resolution_label=1km),
            BNGReference(bng_ref_formatted=SU 13 33, resolution_label=1km),
            BNGReference(bng_ref_formatted=SU 11 34, resolution_label=1km),
            BNGReference(bng_ref_formatted=SU 12 34, resolution_label=1km),
            BNGReference(bng_ref_formatted=SU 13 34, resolution_label=1km),
            BNGReference(bng_ref_formatted=SU 11 35, resolution_label=1km),
            BNGReference(bng_ref_formatted=SU 12 35, resolution_label=1km),
            BNGReference(bng_ref_formatted=SU 13 35, resolution_label=1km)
            ]
            >>> BNGReference("SU1234").bng_dwithin(1001)
            [list of 21 BNGReference objects]

        See Also:
            The equivalent :func:`osbng.traversal.bng_dwithin` function.
        """
        from osbng.traversal import bng_dwithin as _bng_dwithin

        return _bng_dwithin(self, d)


def _validate_bngreferences(func: Callable) -> Callable:
    """Validates that a BNGReference object is passed as an arg or kwarg as expected."""

    @wraps(func)
    def wrapper(*args: Union[BNGReference], **kwargs: Union[BNGReference]) -> Callable:
        # Get the function's signature
        signature = inspect.signature(func)

        # Construct a bound form of the signature
        bound_signature = signature.bind(*args, **kwargs)

        # Iterate through each parameter in the signature
        for arg_name in signature.parameters.keys():
            # Identify the expected data type
            expected_type = signature.parameters.get(arg_name).annotation

            # Find the actual object provided to the argument
            arg_val = bound_signature.arguments.get(arg_name)

            # If a BNGReference is expected and the arg value is not a BNGReference,
            # raise an error
            if (expected_type == BNGReference) and not isinstance(
                arg_val, BNGReference
            ):
                raise TypeError(
                    f"A BNGReference object must be provided as the{arg_name} argument."
                )

        return func(*args, **kwargs)

    return wrapper



"""Index geometries in a ``GeoPandas`` ``GeoDataFrame`` against the BNG index system.

Note:
    This module requires the `GeoPandas <https://github.com/geopandas/geopandas>`__
    package to be installed.

    To install the required package, use:

        pip install osbng[geopandas]

"""

from shapely import Geometry

try:
    import geopandas as gpd
except ImportError as e:
    raise ImportError(
        "The geopandas package is required to use the osbng.indexing_gdf module."
        "Install it with: pip install osbng[geopandas]"
    ) from e

from osbng.indexing import (
    _validate_and_normalise_bng_resolution,
    geom_to_bng_intersection,
)

__all__ = ["gdf_to_bng_intersection_explode"]


def _is_geometry_column(gdf: gpd.GeoDataFrame, col: str) -> bool:
    """Checks if a column in a ``GeoDataFrame`` is a geometry column.

    Args:
        gdf (gpd.GeoDataFrame): The ``GeoDataFrame`` to check.
        col (str): The column name to check.

    Returns:
        bool: True if the column is a geometry column, False otherwise.
    """
    # Drop NULL rows
    non_null = gdf[col].dropna()

    # If the column is empty, return False
    if non_null.empty:
        return False
    # Check if the first non-null value is an instance of Geometry
    return isinstance(non_null.iloc[0], Geometry)


def gdf_to_bng_intersection_explode(
    gdf: gpd.GeoDataFrame,
    resolution: int | str,
    *,
    reset_index: bool = True,
) -> gpd.GeoDataFrame:
    """Applies ``geom_to_bng_intersection`` to a ``GeoDataFrame`` at a given resolution.

    Decomposes each geometry in the input ``GeoDataFrame`` bounded by their presence in
    grid squares at the specified resolution. The resulting
    :class:`~osbng.indexing.BNGIndexedGeometry` objects are exploded into individual
    rows, with each row containing a new column for each
    :class:`~osbng.indexing.BNGIndexedGeometry` object property: ``bng_ref``,
    ``is_core``, and ``geom``.

    Notes:
        Decomposition is achieved by applying the
        :class:`~osbng.indexing.geom_to_bng_intersection` function to each geometry in
        the input ``GeoPandas`` ``GeoDataFrame``, returning a flattened ``GeoDataFrame``
        by exploding the resulting :class:`~osbng.indexing.BNGIndexedGeometry` lists.

        The input ``GeoDataFrame`` geometry column is replaced with the geom property of
        the :class:`~osbng.indexing.BNGIndexedGeometry` objects. The input geometry
        column can be retrieved if required by joining the resulting ``GeoDataFrame``
        with the original ``GeoDataFrame`` on the index (if not reset), or using a
        feature identifier. Dropping the original geometry column reduces memory usage
        and simplifies the resulting ``GeoDataFrame``.

        All non-geometry columns from the original ``GeoDataFrame`` are retained in the
        resulting ``GeoDataFrame``.

        Exploding the resulting ``GeoDataFrame`` allows for easier analysis and
        manipulation of the :class:`~osbng.indexing.BNGIndexedGeometry` object
        properties. This is otherwise a more complex operation.

    Warnings:
        The active geometry column of the input ``GeoDataFrame`` is passed to
        :class:`~osbng.indexing.geom_to_bng_intersection`, which is expected to be set
        and in the OSGB36 / British National Grid coordinate reference system (CRS)
        (EPSG:27700).

    Args:
        gdf (gpd.GeoDataFrame): Input ``GeoPandas`` ``GeoDataFrame``.
        resolution (int | str): The BNG resolution expressed either as a metre-based
            integer or as a string label.

    Keyword Args:
        reset_index (bool): Whether to reset the index of the resulting
          ``GeoDataFrame``, defaults to True.

    Returns:
        gpd.GeoDataFrame: A new ``GeoDataFrame`` with one row per
        :class:`~osbng.indexing.BNGIndexedGeometry` object, containing three columns
        bng_ref, is_core, and geometry corresponding to the
        :class:`~osbng.indexing.BNGIndexedGeometry` object properties.

    Raises:
        BNGResolutionError: If an invalid resolution is provided.
        BNGExtentError: If the coordinates of a Point geometry are outside of the BNG
            index system extent.
        TypeError: If the input is not a ``GeoPandas`` ``GeoDataFrame``.
        ValueError: If the ``GeoDataFrame`` CRS is not equal to "EPSG:27700"
        ValueError: If an active geometry column is not set in the ``GeoDataFrame``.
        ValueError: If the geometry type is not supported.

    Examples:
        >>> import geopandas as gpd
        >>> import pandas as pd
        >>> df = pd.DataFrame({
        ...             "id": ["0", "1", "2"],
        ...             "geometry": [
        ...                 "POLYGON((530000 180000, 535000 180000, 535000 185000,
                             530000 185000, 530000 180000))",
        ...                 "POLYGON((540000 190000, 545000 190000, 545000 195000,
                            540000 195000, 540000 190000))",
        ...                 "POLYGON((550000 200000, 555000 200000, 555000 205000,
                            550000 205000, 550000 200000))"
        ...             ]
        ... })
        >>> gdf = gpd.GeoDataFrame(df, geometry=gpd.GeoSeries.from_wkt(df["geometry"]),
                                   crs="EPSG:27700")
        >>> gdf_to_bng_intersection_explode(gdf, resolution="1km")
            id                                            bng_ref
        0   0  BNGReference(bng_ref_formatted=TQ 31 80, resol...
        1   0  BNGReference(bng_ref_formatted=TQ 30 82, resol...
        2   0  BNGReference(bng_ref_formatted=TQ 33 84, resol...
        3   0  BNGReference(bng_ref_formatted=TQ 34 80, resol...
        4   0  BNGReference(bng_ref_formatted=TQ 33 80, resol...
        ..  ...                                                ...
        70  2  BNGReference(bng_ref_formatted=TL 52 01, resol...
        71  2  BNGReference(bng_ref_formatted=TL 52 03, resol...
        72  2  BNGReference(bng_ref_formatted=TL 50 04, resol...
        73  2  BNGReference(bng_ref_formatted=TL 50 03, resol...
        74  2  BNGReference(bng_ref_formatted=TL 54 01, resol...
            is_core                                       geometry
        0   True  POLYGON ((532000 180000, 532000 181000, 531000...
        1   True  POLYGON ((531000 182000, 531000 183000, 530000...
        2   True  POLYGON ((534000 184000, 534000 185000, 533000...
        3   True  POLYGON ((535000 180000, 535000 181000, 534000...
        4   True  POLYGON ((534000 180000, 534000 181000, 533000...
        ..   ...                                                ...
        70  True  POLYGON ((553000 201000, 553000 202000, 552000...
        71  True  POLYGON ((553000 203000, 553000 204000, 552000...
        72  True  POLYGON ((551000 204000, 551000 205000, 550000...
        73  True  POLYGON ((551000 203000, 551000 204000, 550000...
        74  True  POLYGON ((555000 201000, 555000 202000, 554000...
        [75 rows x 4 columns]

    See Also:
        :class:`~osbng.indexing.geom_to_bng_intersection` to index ``Shapely``
        geometries directly.

    """
    # Validate and normalise the resolution to its metre-based integer value
    validated_resolution = _validate_and_normalise_bng_resolution(resolution)

    # Validate the input is a GeoDataFrame
    if not isinstance(gdf, gpd.GeoDataFrame):
        raise TypeError("Input must be a GeoPandas GeoDataFrame.")

    # Validate the GeoDataFrame is not empty
    if gdf.empty:
        raise ValueError("Input GeoDataFrame must not be empty.")

    # Validate the GeoDataFrame coordinate reference system (CRS) is equal to EPSG:27700
    if gdf.crs is None or not gdf.crs.to_epsg() == 27700:
        raise ValueError(
            "GeoDataFrame CRS must be set to 'EPSG:27700' (British National Grid)."
        )

    # Validate if an active geometry column has been set on the GeoDataFrame
    geometry_column = gdf.active_geometry_name

    if geometry_column is None:
        raise ValueError(
            "GeoDataFrame must have an active geometry column set. "
            "Use gdf.set_geometry(geometry_column_name) to set the active geometry "
            "column."
        )

    # Retain only the active geometry column and non-geometry columns
    gdf = gdf[
        [geometry_column]
        + [col for col in gdf.columns if not _is_geometry_column(gdf, col)]
    ]

    # Initialise an empty list to store the rows for the new GeoDataFrame
    rows = []

    # Iterate over each row in the GeoDataFrame
    for idx, row in gdf.iterrows():
        # Extract the geometry from the specified geometry column
        geom = row[geometry_column]
        # Convert the geometry to BNGIndexedGeometry objects
        bng_idx_geoms = geom_to_bng_intersection(geom, validated_resolution)
        # Drop the geometry column from the original row
        orig_row = row.drop(geometry_column)

        # Iterate over BNGIndexedGeometry objects and create a new row for each
        for bng_idx_geom in bng_idx_geoms:
            # Copy original row columns to a new dictionary
            out_row = orig_row.to_dict()
            # Update the new row with BNGIndexedGeometry properties
            # Retain the original GeoDataFrame index for reference
            out_row.update(
                {
                    "bng_ref": bng_idx_geom.bng_ref,
                    "is_core": bng_idx_geom.is_core,
                    "geometry": bng_idx_geom.geom,
                    "orig_index": idx,
                }
            )
            # Append the new row to the list of rows
            rows.append(out_row)

    # Create a new GeoDataFrame from the list of rows
    out_gdf = gpd.GeoDataFrame(rows, geometry="geometry", crs=27700)

    if reset_index:
        # Drop the orig_index column if reset_index is True
        out_gdf = out_gdf.drop(columns=["orig_index"])
    else:
        # If reset_index is False, set the orig_index column
        # as the index of the result GeoDataFrame
        out_gdf = out_gdf.set_index("orig_index")
        # Set GeoDataFrame to have an unamed index
        out_gdf.index.name = None

    return out_gdf



"""Testing for the hierarchy module.

The test cases are defined in the JSON file located at ./data/hierarchy_test_cases.json
and are used to parameterise the tests for various functions in the indexing module.
Test cases are loaded from the JSON file using the _load_test_cases function, which is
defined in the utils module. The test cases are defined as TypedDicts, which provide a
way to define the structure of the test case data.
"""

from pathlib import Path
from typing import TypedDict

import pytest

from osbng.bng_reference import BNGReference
from osbng.errors import _EXCEPTION_MAP
from osbng.hierarchy import bng_to_children, bng_to_parent
from osbng.utils import _load_test_cases


class BNGToChildrenTestCase(TypedDict):
    """TypedDict for bng_to_children test cases.

    Attributes:
        bng_ref_string (str): The BNG reference string.
        resolution (int | str): The resolution expressed either as a metre-based integer
            or as a string label.
        expected_exception (dict[str, str] | None): The expected exception is a
            dictionary with the exception name and optional message.
        expected (dict[str, list[str]] | None): The expected result is a dictionary with
            a list of BNG reference formatted strings.
    """

    bng_ref_string: str
    resolution: int | str
    expected_exception: dict[str, str] | None
    expected: dict[str, list[str]] | None


# Parameterised test for bng_to_children function
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "hierarchy_test_cases.json"
    )["bng_to_children"],
)
def test_bng_to_children(test_case: BNGToChildrenTestCase) -> None:
    """Test bng_to_children with test cases from JSON file.

    Args:
        test_case (BNGToChildrenTestCase): Test case from JSON file.
    """
    # Load test case data
    bng_ref_string = test_case["bng_ref_string"]
    resolution = None if test_case["resolution"] == "NULL" else test_case["resolution"]
    # Get expected result
    expected = (
        None
        if "expected_exception" in test_case
        else test_case["expected"]["bng_ref_formatted"]
    )

    if "expected_exception" in test_case:
        # Get exception name from test case
        exception_name = test_case["expected_exception"]["name"]
        # Get exception message from test case
        message = (
            None
            if test_case["expected_exception"]["name"] == "BNGResolutionError"
            else test_case["expected_exception"]["message"]
        )
        # Get exception class from name
        exception_class = _EXCEPTION_MAP[exception_name]
        # Assert that the test case raises the expected exception and message
        with pytest.raises(exception_class, match=message):
            bng_to_children(BNGReference(bng_ref_string), resolution=resolution)

    else:
        # Return a list of child BNGReference objects
        bng_refs = bng_to_children(BNGReference(bng_ref_string), resolution=resolution)
        # Sort lists to account for order differences
        bng_ref_strings = [bng_ref.bng_ref_formatted for bng_ref in bng_refs]
        # Assert that the function returns the expected result
        assert sorted(bng_ref_strings) == sorted(expected)


class BNGToParentTestCase(TypedDict):
    """TypedDict for bng_to_parent test cases.

    Attributes:
        bng_ref_string (str): The BNG reference string.
        resolution (int | str): The resolution expressed either as a metre-based integer
            or as a string label.
        expected_exception (dict[str, str] | None): The expected exception is a
            dictionary with the exception name.
        expected (dict[str, str] | None): The expected result is a dictionary with the
            BNG reference formatted string.
    """

    bng_ref_string: str
    resolution: int | str
    expected_exception: dict[str, str] | None
    expected: dict[str, str] | None


# Parameterised test for bng_to_parent function
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "hierarchy_test_cases.json"
    )["bng_to_parent"],
)
def test_bng_to_parent(test_case: BNGToParentTestCase) -> None:
    """Test bng_to_parent with test cases from JSON file.

    Args:
        test_case (BNGToParentTestCase): Test case from JSON file.
    """
    # Load test case data
    bng_ref_string = test_case["bng_ref_string"]
    resolution = None if test_case["resolution"] == "NULL" else test_case["resolution"]

    if "expected_exception" in test_case:
        # Get exception name from test case
        exception_name = test_case["expected_exception"]["name"]
        # Get exception message from test case
        message = (
            None
            if test_case["expected_exception"]["name"] == "BNGResolutionError"
            else test_case["expected_exception"]["message"]
        )
        # Get exception class from name
        exception_class = _EXCEPTION_MAP[exception_name]
        # Assert that the test case raises the expected exception and message
        with pytest.raises(exception_class, match=message):
            bng_to_parent(BNGReference(bng_ref_string), resolution=resolution)

    else:
        # Return the parent BNGReference object
        bng_ref = bng_to_parent(BNGReference(bng_ref_string), resolution=resolution)
        # Assert that the function returns the expected result
        assert bng_ref.bng_ref_formatted == test_case["expected"]["bng_ref_formatted"]



"""Testing for the traversal module.

The test cases are defined in the JSON file located at ./data/traversal_test_cases.json
and are used to parameterise the tests for various functions in the traversal module.
Test cases are loaded from the JSON file using the _load_test_cases function, which is
defined in the utils module. The test cases are defined as TypedDicts, which provide a
way to define the structure of the test case data.
"""

from pathlib import Path
from typing import TypedDict

import pytest

from osbng.bng_reference import BNGReference
from osbng.errors import _EXCEPTION_MAP
from osbng.traversal import (
    bng_distance,
    bng_dwithin,
    bng_is_neighbour,
    bng_kdisc,
    bng_kring,
)
from osbng.utils import _load_test_cases


class BNGDistanceTestCase(TypedDict):
    """TypedDict for bng_distance function test cases.

    Attributes:
        bng_ref_string_1 (str): The first BNG reference string.
        bng_ref_string_2 (str): The second BNG reference string.
        edge_to_edge (bool | None): Whether to calculate edge-to-edge distance.
        expected (float): The expected distance between the pair of BNGReference
            objects.
    """

    bng_ref_string_1: str
    bng_ref_string_2: str
    edge_to_edge: bool | None
    expected: float


# Parameterised test for bng_distance function
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "traversal_test_cases.json"
    )["bng_distance"],
)
def test_bng_distance(test_case: BNGDistanceTestCase) -> None:
    """Test bng_distance with test cases from JSON file.

    Args:
        test_case (BNGDistanceTestCase): The test case from JSON file.
    """
    # Load test case data
    bng_ref1 = test_case["bng_ref_string_1"]
    bng_ref2 = test_case["bng_ref_string_2"]

    if "expected_exception" in test_case:
        # Get exception name from test case
        exception_name = test_case["expected_exception"]["name"]
        # Get exception class from name
        exception_class = _EXCEPTION_MAP[exception_name]
        # Assert that the test case raises the expected exception
        with pytest.raises(exception_class):
            bng_distance(BNGReference(bng_ref1), BNGReference(bng_ref2))

    elif "edge_to_edge" in test_case:
        # If edge_to_edge is specified, use it in the distance calculation
        edge_to_edge = test_case["edge_to_edge"]
        # Assert that the function returns the expected result
        distance = bng_distance(
            BNGReference(bng_ref1), BNGReference(bng_ref2), edge_to_edge=edge_to_edge
        )
        assert distance == test_case["expected"]

    else:
        # Assert that the function returns the expected result
        distance = bng_distance(BNGReference(bng_ref1), BNGReference(bng_ref2))
        assert distance == test_case["expected"]


class BNGIsNeighbourTestCase(TypedDict):
    """TypedDict for bng_is_neighbour function test cases.

    Attributes:
        bng_ref_string_1 (str): The first BNG reference string.
        bng_ref_string_2 (str): The second BNG reference string.
        expected_exception (dict[str, str] | None): The expected exception is a
            dictionary with the exception name and message.
        expected (bool | None): The expected result of the neighbour check.
    """

    bng_ref_string_1: str
    bng_ref_string_2: str
    expected_exception: dict[str, str] | None
    expected: bool | None


# Parameterised test for bng_is_neighbour function
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "traversal_test_cases.json"
    )["bng_is_neighbour"],
)
def test_bng_is_neighbour(test_case: BNGIsNeighbourTestCase) -> None:
    """Test bng_is_neighbour with test cases from JSON file.

    Args:
        test_case (BNGIsNeighbourTestCase): The test case from JSON file.
    """
    # Load test case data
    bng_ref1 = test_case["bng_ref_string_1"]
    bng_ref2 = test_case["bng_ref_string_2"]

    if "expected_exception" in test_case:
        # Get exception name from test case
        exception_name = test_case["expected_exception"]["name"]
        # Get exception message from test case
        message = (
            None
            if test_case["expected_exception"]["name"] == "BNGNeighbourError"
            else test_case["expected_exception"]["message"]
        )
        # Get exception class from name
        exception_class = _EXCEPTION_MAP[exception_name]
        # Assert that the test case raises the expected exception
        with pytest.raises(exception_class, match=message):
            bng_is_neighbour(BNGReference(bng_ref1), BNGReference(bng_ref2))

    else:
        # Assert that the function returns the expected result
        distance = bng_is_neighbour(BNGReference(bng_ref1), BNGReference(bng_ref2))
        assert distance == test_case["expected"]


class BNGKRingTestCase(TypedDict):
    """TypedDict for bng_kring function test cases.

    Attributes:
        bng_ref_string (str): The BNG reference string.
        k (int): The k value for the k-ring.
        expected_warning (bool | None): The expected warning is a boolean indicating if
            a warning is expected.
        expected (dict[str, list[str]] | None): The expected result is a dictionary with
            the key "bng_ref_formatted" and a list of formatted BNG reference strings.
        expected_length (int | None): The expected length of the k-ring. Represents the
            number of BNGReference objects within k-ring.
    """

    bng_ref_string: str
    k: int
    expected_warning: bool | None
    expected: dict[str, list[str]] | None
    expected_length: int | None


# Parameterised test for bng_kring function
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "traversal_test_cases.json"
    )["bng_kring"],
)
def test_bng_kring(test_case: BNGKRingTestCase) -> None:
    """Test bng_kring with test cases from JSON file.

    Args:
        test_case (BNGKRingTestCase): The test case from JSON file.
    """
    # Load test case data
    bng_ref_string = test_case["bng_ref_string"]
    k = test_case["k"]
    # Get expected result
    expected = (
        None
        if "expected_length" in test_case
        else test_case["expected"]["bng_ref_formatted"]
    )
    expected_length = None if "expected" in test_case else test_case["expected_length"]

    if "expected_length" in test_case:
        # Assert that the function returns the expected length
        assert len(bng_kring(BNGReference(bng_ref_string), k)) == expected_length

    elif "expected_warning" in test_case:
        # Assert that the test case raises a warning
        with pytest.warns(UserWarning):
            # Assert that the function returns the expected result
            kring = bng_kring(BNGReference(bng_ref_string), k)
            assert sorted([r.bng_ref_formatted for r in kring]) == sorted(expected)

    else:
        # Assert that the function returns the expected result
        kring = bng_kring(BNGReference(bng_ref_string), k)
        assert sorted([r.bng_ref_formatted for r in kring]) == sorted(expected)


class BNGKDiscTestCase(TypedDict):
    """TypedDict for bng_kdisc function test cases.

    Attributes:
        bng_ref_string (str): The BNG reference string.
        k (int): The k value for the k-disc.
        expected_warning (bool | None): The expected warning is a boolean indicating if
            a warning is expected.
        expected (dict[str, list[str]] | None): The expected result is a dictionary with
            the key "bng_ref_formatted" and a list of formatted BNG reference strings.
        expected_length (int | None): The expected length of the k-disc. Represents the
            number of BNGReference objects within k-disc.
    """

    bng_ref_string: str
    k: int
    expected_warning: bool | None
    expected: dict[str, list[str]] | None
    expected_length: int | None


# Parameterised test for bng_kdisc function
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "traversal_test_cases.json"
    )["bng_kdisc"],
)
def test_bng_kdisc(test_case: BNGKDiscTestCase) -> None:
    """Test bng_kdisc with test cases from JSON file.

    Args:
        test_case (BNGKDiscTestCase): The test case from JSON file.
    """
    # Load test case data
    bng_ref_string = test_case["bng_ref_string"]
    k = test_case["k"]
    # Get expected result
    expected = (
        None
        if "expected_length" in test_case
        else test_case["expected"]["bng_ref_formatted"]
    )
    expected_length = None if "expected" in test_case else test_case["expected_length"]

    if "expected_length" in test_case:
        # Assert that the function returns the expected length
        assert len(bng_kdisc(BNGReference(bng_ref_string), k)) == expected_length

    elif "expected_warning" in test_case:
        # Assert that the test case raises a warning
        with pytest.warns(UserWarning):
            # Assert that the function returns the expected result
            kdisc = bng_kdisc(BNGReference(bng_ref_string), k)
            assert sorted([r.bng_ref_formatted for r in kdisc]) == sorted(expected)

    else:
        # Assert that the function returns the expected result
        kdisc = bng_kdisc(BNGReference(bng_ref_string), k)
        assert sorted([r.bng_ref_formatted for r in kdisc]) == sorted(expected)


class BNGKDWithinTestCase(TypedDict):
    """TypedDict for bng_dwithin function test cases.

    Attributes:
        bng_ref_string (str): The BNG reference string.
        d (int): The d value for the d-within search.
        expected (dict[str, list[str]] | None): The expected result is a dictionary with
            the key "bng_ref_formatted" and a list of formatted BNG reference strings.
        expected_length (int | None): The expected length of the d-within search.
            Represents the number of BNGReference objects within d-within search.
    """

    bng_ref_string: str
    d: int
    expected: dict[str, list[str]] | None
    expected_length: int | None


# Parameterised test for bng_dwithin function
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "traversal_test_cases.json"
    )["bng_dwithin"],
)
def test_bng_dwithin(test_case: BNGKDWithinTestCase) -> None:
    """Test bng_dwithin with test cases from JSON file.

    Args:
        test_case (BNGKDWithinTestCase): The test case from JSON file.
    """
    if "expected_length" in test_case:
        # Assert that the function returns the expected length
        assert (
            len(bng_dwithin(BNGReference(test_case["bng_ref_string"]), test_case["d"]))
            == test_case["expected_length"]
        )
    else:
        # Assert that the function returns the expected result
        kring = bng_dwithin(BNGReference(test_case["bng_ref_string"]), test_case["d"])
        assert sorted([r.bng_ref_formatted for r in kring]) == sorted(
            test_case["expected"]["bng_ref_formatted"]
        )



"""Navigate the hierarchical structure of the BNG index system.

The British National Grid (BNG) is structured using a hierarchical system of grid
squares at various resolutions. At its highest level, the grid divides GB into 100 km by
100 km squares, each identified by a two-letter code. Successive levels of resolution
further subdivide the grid squares into finer detail, down to individual 1-meter
squares. This module allows for the traversal of this hierarchy by providing methods to
return the parent and children of :class:`~osbng.bng_reference.BNGReference` objects at
specified resolutions.

Parent and child definitions:
    - **Parent**: The parent of a :class:`~osbng.bng_reference.BNGReference` object is
      the grid square at the next higher (coarser) resolution level that contains the
      current reference. For example, the parent of a 1km grid square reference would be
      the 5km grid square that contains it.
    - **Children**: The children of a :class:`~osbng.bng_reference.BNGReference` object
      are the grid squares at the next lower (finer) resolution level that are contained
      within the current reference. For example, the children of a 10km grid square
      reference would be the 5km grid squares that it contains.

Notes:
    While parent and child derivation defaults to the next higher and lower
    resolution, any supported resolution in the hierarchy can be specified.

Supported Resolutions:
    - The module supports the 'standard' and 'intermediate' quadtree resolutions,
      including ``100km``, ``50km``, ``10km``, ``5km``, ``1km``, ``500m``, ``100m``,
      ``50m``, ``10m``, ``5m`` and ``1m``.
    - These resolutions passed to hierarchy functions are validated and normalised
      using the resolution mapping defined in the :doc:`resolution` module.
"""

from osbng.bng_reference import BNGReference, _validate_bngreferences
from osbng.errors import BNGHierarchyError
from osbng.indexing import (
    _validate_and_normalise_bng_resolution,
    bbox_to_bng,
    bng_to_xy,
    xy_to_bng,
)
from osbng.resolution import BNG_RESOLUTIONS

__all__ = ["bng_to_children", "bng_to_parent"]


@_validate_bngreferences
def bng_to_children(
    bng_ref: BNGReference, *, resolution: int | str | None = None
) -> list[BNGReference]:
    """Returns a list of child ``BNGReference`` objects of a ``BNGReference``.

    By default, the children of the :class:`~osbng.bng_reference.BNGReference` object is
    defined as the :class:`~osbng.bng_reference.BNGReference` objects in the next
    resolution down from the input :class:`~osbng.bng_reference.BNGReference` resolution
    . For example, 100km -> 50km.

    Notes:
        Any valid resolution can be provided as the child resolution, provided it is
        less than the resolution of the input :class:`~osbng.bng_reference.BNGReference`
        .

    Args:
        bng_ref (BNGReference): The :class:`~osbng.bng_reference.BNGReference` object to
            derive children from.

    Keyword Args:
        resolution (int | str | None, optional): The resolution of
            the children :class:`~osbng.bng_reference.BNGReference` objects expressed
            either as a metre-based integer or as a string label. Defaults to None.

    Returns:
        list[BNGReference]: A list of :class:`~osbng.bng_reference.BNGReference` objects
        that are children of the input :class:`~osbng.bng_reference.BNGReference` object
        .

    Raises:
        BNGReferenceError: If the first positional argument is not a
            :class:`~osbng.bng_reference.BNGReference` object.
        BNGHierarchyError: If the resolution of the input
            :class:`~osbng.bng_reference.BNGReference` object is 1m.
        BNGHierarchyError: If the resolution is greater than or equal to the resolution
            of the input :class:`~osbng.bng_reference.BNGReference` object.
        BNGResolutionError: If an invalid resolution is provided.

    Examples:
        >>> bng_to_children(BNGReference("SU"))
        [BNGReference(bng_ref_formatted=SU SW, resolution_label=50km),
        BNGReference(bng_ref_formatted=SU SE, resolution_label=50km),
        BNGReference(bng_ref_formatted=SU NW, resolution_label=50km),
        BNGReference(bng_ref_formatted=SU NE, resolution_label=50km)]
        >>> bng_to_children(BNGReference("SU36"))
        [BNGReference(bng_ref_formatted=SU 3 6 SW, resolution_label=5km),
        BNGReference(bng_ref_formatted=SU 3 6 SE, resolution_label=5km),
        BNGReference(bng_ref_formatted=SU 3 6 NW, resolution_label=5km),
        BNGReference(bng_ref_formatted=SU 3 6 NE, resolution_label=5km)]

    See Also:
        - The equivalent :meth:`osbng.bng_reference.BNGReference.bng_to_children`
          instance method.
        - :class:`~osbng.hierarchy.bng_to_parent` to derive the
          :class:`~osbng.bng_reference.BNGReference` object in the next resolution up.

    """
    # Raise error if the resolution is 1m
    if bng_ref.resolution_metres == 1:
        raise BNGHierarchyError("Cannot derive children from the finest 1m resolution")

    # Generate a scaled resolution if none is provided
    if resolution is None:
        resolution = bng_ref.resolution_metres

        if BNG_RESOLUTIONS[resolution]["quadtree"]:
            resolution = int((bng_ref.resolution_metres / 5))
        else:
            resolution = int((bng_ref.resolution_metres / 2))

    # Validate and normalise the resolution to its metre-based integer value
    validated_resolution = _validate_and_normalise_bng_resolution(resolution)

    # Raise error if the validated resolution is greater than the resolution of the
    # input BNGReference object
    if validated_resolution >= bng_ref.resolution_metres:
        raise BNGHierarchyError(
            "Resolution must be less than the resolution of input BNGReference object"
        )

    # Get min and max coordinates of the grid square bounding box
    min_coords = bng_to_xy(bng_ref, position="lower-left")
    max_coords = bng_to_xy(bng_ref, position="upper-right")

    # Derive children BNGReference objects from the bounding box
    bng_refs = bbox_to_bng(
        min_coords[0], min_coords[1], max_coords[0], max_coords[1], validated_resolution
    )

    return bng_refs


@_validate_bngreferences
def bng_to_parent(
    bng_ref: BNGReference, *, resolution: int | str | None = None
) -> BNGReference:
    """Returns the `BNGReference`that is the parent of a ``BNGReference``.

    By default, the parent of the :class:`~osbng.bng_reference.BNGReference` object is
    defined as the :class:`~osbng.bng_reference.BNGReference` in the next BNG resolution
    up from the input :class:`~osbng.bng_reference.BNGReference` resolution. For example
    , 50km -> 100km.

    Notes:
        Any valid resolution can be provided as the parent resolution, provided it is
        greater than the resolution of the input
        :class:`~osbng.bng_reference.BNGReference`.

    Args:
        bng_ref (BNGReference): The :class:`~osbng.bng_reference.BNGReference` object to
            derive parent from.

    Keyword Args:
        resolution (int | str | None, optional): The resolution of the parent
            :class:`~osbng.bng_reference.BNGReference` objects expressed either as a
            metre-based integer or as a string label. Defaults to None.

    Returns:
        BNGReference: A :class:`~osbng.bng_reference.BNGReference` object that is the
        parent of the input :class:`~osbng.bng_reference.BNGReference` object.

    Raises:
        BNGReferenceError: If the first positional argument is not a
            :class:`~osbng.bng_reference.BNGReference` object.
        BNGHierarchyError: If the resolution of the input
            :class:`~osbng.bng_reference.BNGReference` object is 100km.
        BNGHierarchyError: If the resolution is less than or equal to the resolution of
            the input :class:`~osbng.bng_reference.BNGReference` object.
        BNGResolutionError: If an invalid resolution is provided.

    Examples:
        >>> bng_to_parent(BNGReference("SU 3 6 SW"))
        BNGReference(bng_ref_formatted=SU 3 6, resolution_label=10km)
        >>> bng_to_parent(BNGReference("SU 342 567"))
        BNGReference(bng_ref_formatted=SU 34 56 NW, resolution_label=500m)
        >>> bng_to_parent(BNGReference("SU 342 567"), resolution=10000)
        BNGReference(bng_ref_formatted=SU 3 5, resolution_label=10km)

    See Also:
        - The equivalent :meth:`osbng.bng_reference.BNGReference.bng_to_parent`
          instance method.
        - :class:`~osbng.hierarchy.bng_to_children` to derive the
          :class:`~osbng.bng_reference.BNGReference` objects in the next resolution
          down.

    """
    # Raise error if the resolution is 100km
    if bng_ref.resolution_metres == 100000:
        raise BNGHierarchyError(
            "Cannot derive parent from the coarsest 100km resolution"
        )

    # Generate a scaled resolution if none is provided
    if resolution is None:
        resolution = bng_ref.resolution_metres

        if BNG_RESOLUTIONS[resolution]["quadtree"]:
            resolution = int((bng_ref.resolution_metres * 2))
        else:
            resolution = int((bng_ref.resolution_metres * 5))

    # Validate and normalise the resolution to its metre-based integer value
    validated_resolution = _validate_and_normalise_bng_resolution(resolution)

    # Raise error if the validated resolution is less than the resolution of the input
    # BNGReference object
    if validated_resolution <= bng_ref.resolution_metres:
        raise BNGHierarchyError(
            "Resolution must be greater than the resolution of input BNGReference "
            "object"
        )

    # Dervive coordinates of the grid square bounding box
    x, y = bng_to_xy(bng_ref, position="lower-left")

    # Derive parent BNGReference object from coordinates
    bng_ref = xy_to_bng(x, y, validated_resolution)

    return bng_ref



"""Geospatial grid indexing and interaction with the British National Grid.

Offers a streamlined programmatic interface to Ordnance Survey's British National Grid
(BNG) index system, enabling efficient spatial indexing and analysis based on grid
references. It supports a range of geospatial applications, including statistical
aggregation, data visualisation, and interoperability across datasets. Designed for
developers and analysts working with geospatial data in Great Britain, osbng simplifies
integration with geospatial workflows and provides intuitive tools for exploring the
structure and logic of the BNG system.
"""

from osbng.bng_reference import BNGReference
from osbng.grids import (
    BNG_BOUNDS,
    bbox_to_bng_iterfeatures,
    bng_grid_1km,
    bng_grid_5km,
    bng_grid_10km,
    bng_grid_50km,
    bng_grid_100km,
)
from osbng.indexing import (
    PREFIXES,
    SUFFIXES,
    bbox_to_bng,
    geom_to_bng,
    geom_to_bng_intersection,
    xy_to_bng,
)
from osbng.resolution import BNG_RESOLUTIONS

__all__ = [
    "BNG_BOUNDS",
    "BNG_RESOLUTIONS",
    "PREFIXES",
    "SUFFIXES",
    "BNGReference",
    "bng_grid_100km",
    "bng_grid_50km",
    "bng_grid_10km",
    "bng_grid_5km",
    "bng_grid_1km",
    "xy_to_bng",
    "bbox_to_bng",
    "bbox_to_bng_iterfeatures",
    "geom_to_bng",
    "geom_to_bng_intersection",
]
__version__ = "0.5.1"



"""Generate British National Grid (BNG) grid square data within specified bounds.

Uses a GeoJSON-like mapping for grid squares implementing the `__geo_interface__
<https://gist.github.com/sgillies/2217756>`__. Use of this protocol enables
integration with geospatial data processing libraries and tools.

Grid square data covering the BNG index system bounds is provided as an iterator at
100km, 50km, 10km, 5km and 1km resolutions. ``GeoPandas`` can be used to read the
iterator data directly into a ``GeoDataFrame`` for further processing using
`geopandas.GeoDataFrame.from_features()
<https://geopandas.org/en/stable/docs/reference/api/geopandas.GeoDataFrame.from_features
.html>`__ or similar methods. Iterators can be converted to lists to generate all grid
square GeoJSON-like Features at a given resolution.
"""

from typing import Any, Iterator

from osbng.indexing import bbox_to_bng

__all__ = [
    "BNG_BOUNDS",
    "bbox_to_bng_iterfeatures",
    "bng_grid_100km",
    "bng_grid_50km",
    "bng_grid_10km",
    "bng_grid_5km",
    "bng_grid_1km",
]

# BNG index system bounds
BNG_BOUNDS: tuple[int, int, int, int] = (0, 0, 700000, 1300000)
"""tuple[int, int, int, int]: BNG index system bounding box.

Expressed as (xmin, ymin, xmax, ymax) using easting and northing coordinates.

Represents the valid extent for :class:`~osbng.bng_reference.BNGReference` and grid 
generation.
"""


def bbox_to_bng_iterfeatures(
    xmin: int | float,
    ymin: int | float,
    xmax: int | float,
    ymax: int | float,
    resolution: int | str,
) -> Iterator[dict[str, Any]]:
    """Returns an iterator of BNGReference Features given a bounding box and resolution.

    Implements the `__geo_interface__
    <https://gist.github.com/sgillies/2217756>`__ protocol. The returned data structure
    represents the :class:`~osbng.bng_reference.BNGReference` object as a GeoJSON-like
    Feature.

    Args:
        xmin (int | float): The minimum easting coordinate of the bounding box (BBOX).
        ymin (int | float): The minimum northing coordinate of the BBOX.
        xmax (int | float): The maximum easting coordinate of the BBOX.
        ymax (int | float): The maximum northing coordinate of the BBOX.
        resolution (int | str): The BNG resolution expressed either as a metre-based
            integer or as a string label.

    Yields:
        dict[str, Any]: A GeoJSON-like representation of a BNGReference object.

    Raises:
        BNGResolutionError: If the resolution is not a valid resolution.

    Examples:
        >>> print(*bbox_to_bng_iterfeatures(530000, 180000, 535000, 185000, "5km"))
        {'type': 'Feature', 'properties': {'bng_ref': 'TQ38SW'}, 'geometry':
        {'type': 'Polygon', 'coordinates': (((535000.0, 180000.0), (535000.0, 185000.0),
        (530000.0, 185000.0), (530000.0, 180000.0), (535000.0, 180000.0)),)}}
        >>> print(*bbox_to_bng_iterfeatures(530000, 180000, 535000, 185000, "10km"))
        {'type': 'Feature', 'properties': {'bng_ref': 'TQ38'}, 'geometry':
        {'type': 'Polygon', 'coordinates': (((540000.0, 180000.0), (540000.0, 190000.0),
        (530000.0, 190000.0), (530000.0, 180000.0), (540000.0, 180000.0)),)}}

    """
    # Convert the bounding box to BNGReference objects
    bng_refs = bbox_to_bng(xmin, ymin, xmax, ymax, resolution)

    # Yield BNGReference object GeoJSON-like Features
    for bng_ref in bng_refs:
        yield bng_ref.__geo_interface__


# Grid square data covering the BNG index system bounds provided at
# 100km, 50km, 10km, 5km and 1km resolutions as iterators
# Iterators can be converted to a list to trigger generation of
# BNGReference object Features
# Resolution capped at 1km to prevent excessive data generation
# for lower (finer) resolutions

bng_grid_100km: Iterator[dict[str, Any]] = bbox_to_bng_iterfeatures(
    *BNG_BOUNDS, "100km"
)
"""Iterator of GeoJSON-like Features for 100km BNG grid squares within ``BNG_BOUNDS``.

Each yielded dictionary implements the ``__geo_interface__``  for a single 100km BNG 
grid square and corresponds to a :class:`~osbng.bng_reference.BNGReference` at 100 km 
resolution.

Notes:
    - No Features are generated until the iterator is consumed.
    - Convert to a list (``list(bng_grid_100km)``) to trigger Feature generation.
    - Pass directly to ``gpd.GeoDataFrame.from_features`` for ``GeoDataFrame`` 
      construction.
"""

bng_grid_50km: Iterator[dict[str, Any]] = bbox_to_bng_iterfeatures(*BNG_BOUNDS, "50km")
"""Iterator of GeoJSON-like Features for 50km BNG grid squares within ``BNG_BOUNDS``.

Each yielded dictionary implements the ``__geo_interface__``  for a single 50km BNG grid
square and corresponds to a :class:`~osbng.bng_reference.BNGReference` at 50km 
resolution.

Notes:
    - No Features are generated until the iterator is consumed.
    - Convert to a list (``list(bng_grid_50km)``) to trigger Feature generation.
    - Pass directly to ``gpd.GeoDataFrame.from_features`` for ``GeoDataFrame`` 
      construction.
"""

bng_grid_10km: Iterator[dict[str, Any]] = bbox_to_bng_iterfeatures(*BNG_BOUNDS, "10km")
"""Iterator of GeoJSON-like Features for 10km BNG grid squares within ``BNG_BOUNDS``.

Each yielded dictionary implements the ``__geo_interface__``  for a single 10km BNG grid
square and corresponds to a :class:`~osbng.bng_reference.BNGReference` at 10km 
resolution.

Notes:
    - No Features are generated until the iterator is consumed.
    - Convert to a list (``list(bng_grid_10km)``) to trigger Feature generation.
    - Pass directly to ``gpd.GeoDataFrame.from_features`` for ``GeoDataFrame`` 
      construction.
"""

bng_grid_5km: Iterator[dict[str, Any]] = bbox_to_bng_iterfeatures(*BNG_BOUNDS, "5km")
"""Iterator of GeoJSON-like Features for all 5km BNG grid squares within ``BNG_BOUNDS``.

Each yielded dictionary implements the ``__geo_interface__``  for a single 5km BNG grid 
square and corresponds to a :class:`~osbng.bng_reference.BNGReference` at 5km resolution
.

Notes:
    - No Features are generated until the iterator is consumed.
    - Convert to a list (``list(bng_grid_5km)``) to trigger Feature generation.
    - Pass directly to ``gpd.GeoDataFrame.from_features`` for ``GeoDataFrame`` 
      construction.
"""

bng_grid_1km: Iterator[dict[str, Any]] = bbox_to_bng_iterfeatures(*BNG_BOUNDS, "1km")
"""Iterator of GeoJSON-like Features for all 1km BNG grid squares within ``BNG_BOUNDS``.

Each yielded dictionary implements the ``__geo_interface__``  for a single 1km BNG grid 
square and corresponds to a :class:`~osbng.bng_reference.BNGReference` at 1km resolution
.

Notes:
    - No Features are generated until the iterator is consumed.
    - Convert to a list (``list(bng_grid_1km)``) to trigger Feature generation.
    - Pass directly to ``gpd.GeoDataFrame.from_features`` for ``GeoDataFrame`` 
      construction.
"""



"""Testing for the indexing module.

The test cases are defined in the JSON file located at ./data/indexing_test_cases.json
and are used to parameterise the tests for various functions in the indexing module.
Test cases are loaded from the JSON file using the _load_test_cases function, which is
defined in the utils module. The test cases are defined as TypedDicts, which provide a
way to define the structure of the test case data.
"""

from math import sqrt
from pathlib import Path
from typing import Any, TypedDict

import pytest
from shapely import Geometry
from shapely.geometry import shape
from shapely.testing import assert_geometries_equal

from osbng.bng_reference import BNGReference
from osbng.errors import _EXCEPTION_MAP
from osbng.indexing import (
    _decompose_geom,
    _get_bng_suffix,
    _validate_and_normalise_bbox,
    _validate_and_normalise_bng_resolution,
    _validate_easting_northing,
    bbox_to_bng,
    bng_to_bbox,
    bng_to_grid_geom,
    bng_to_xy,
    geom_to_bng,
    geom_to_bng_intersection,
    xy_to_bng,
)
from osbng.utils import _load_test_cases


class ValidateAndNormaliseBNGResolutionTestCase(TypedDict):
    """TypedDict for _validate_and_normalise_bng_resolution function test cases.

    Attributes:
        resolution (int | float | str): The BNG resolution expressed either as a
            metre-based integer or float, or as a string label.
        expected (int | None): The expected result is an integer or None if exception
            is expected.
        expected_exception (dict[str, str] | None): The expected exception is a
            dictionary with the exception name.
    """

    resolution: int | float | str
    expected: int | None
    expected_exception: dict[str, str] | None


# Parameterised test for _validate_and_normalise_bng_resolution function
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "indexing_test_cases.json"
    )["_validate_and_normalise_bng_resolution"],
)
def test__validate_and_normalise_bng_resolution(
    test_case: ValidateAndNormaliseBNGResolutionTestCase,
) -> None:
    """Test _validate_and_normalise_bng_resolution with test cases from JSON file.

    Args:
        test_case (ValidateAndNormaliseBNGResolutionTestCase): Test case from JSON file.
    """
    # Load test case data
    resolution = test_case["resolution"]

    if "expected_exception" in test_case:
        # Get exception name from test case
        exception_name = test_case["expected_exception"]["name"]
        # Get exception class from name
        exception_class = _EXCEPTION_MAP[exception_name]
        # Assert that the test case raises the expected exception
        with pytest.raises(exception_class):
            _validate_and_normalise_bng_resolution(resolution)

    else:
        # Get expected result
        expected = test_case["expected"]
        # Assert that the function returns the expected result
        assert _validate_and_normalise_bng_resolution(resolution) == expected


class ValidateEastingNorthingTestCase(TypedDict):
    """TypedDict for _validate_easting_northing function test cases.

    Attributes:
        easting (int | float): The easting coordinate.
        northing (int | float): The northing coordinate.
        expected_exception (dict[str, str] | None): The expected exception is a
            dictionary with the exception name.
    """

    easting: int | float
    northing: int | float
    expected_exception: dict[str, str] | None


# Parameterised test for _validate_easting_northing function
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "indexing_test_cases.json"
    )["_validate_easting_northing"],
)
def test__validate_easting_northing(test_case: ValidateEastingNorthingTestCase) -> None:
    """Test _validate_and_normalise_bng_resolution with test cases from JSON file.

    Args:
        test_case (ValidateEastingNorthingTestCase): Test case from JSON file.
    """
    # Load test case data
    easting = test_case["easting"]
    northing = test_case["northing"]

    if "expected_exception" in test_case:
        # Get exception name from test case
        exception_name = test_case["expected_exception"]["name"]
        # Get exception class from name
        exception_class = _EXCEPTION_MAP[exception_name]
        # Assert that the test case raises the expected exception
        with pytest.raises(exception_class):
            _validate_easting_northing(easting, northing)

    else:
        # Assert that the function returns the expected result
        _validate_easting_northing(easting, northing)


class ValidateAndNormaliseBBOXTestCase(TypedDict):
    """TypedDict for _validate_and_normalise_bbox function test cases.

    Attributes:
        xmin (int | float): The minimum easting coordinate of the bounding box (BBOX).
        ymin (int | float): The minimum northing coordinate of the BBOX.
        xmax (int | float): The maximum easting coordinate of the BBOX.
        ymax (int | float): The maximum northing coordinate of the BBOX.
        expected_warning (bool | None): The expected warning is a boolean indicating if
            a warning is expected.
        expected (list[int | float]): The expected result is a list of BBOX coordinates.
    """

    xmin: int | float
    ymin: int | float
    xmax: int | float
    ymax: int | float
    expected_warning: bool | None
    expected: list[int | float]


# Parameterised test for _validate_and_normalise_bbox function
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "indexing_test_cases.json"
    )["_validate_and_normalise_bbox"],
)
def test__validate_and_normalise_bbox(
    test_case: ValidateAndNormaliseBBOXTestCase,
) -> None:
    """Test _validate_and_normalise_bbox with test cases from JSON file.

    Args:
        test_case (ValidateAndNormaliseBBOXTestCase): Test case from JSON file.
    """
    # Load test case data
    xmin = test_case["xmin"]
    ymin = test_case["ymin"]
    xmax = test_case["xmax"]
    ymax = test_case["ymax"]
    # Get expected result as tuple
    expected = tuple(test_case["expected"])

    if "expected_warning" in test_case:
        # Assert that the test case raises a warning
        with pytest.warns(UserWarning):
            # Assert that the function returns the expected result
            assert _validate_and_normalise_bbox(xmin, ymin, xmax, ymax) == expected

    else:
        # Assert that the function returns the expected result
        assert _validate_and_normalise_bbox(xmin, ymin, xmax, ymax) == expected


class GetBNGSuffixTestCase(TypedDict):
    """TypedDict for _get_bng_suffix function test cases.

    Attributes:
        easting (int | float): The easting coordinate.
        northing (int | float): The northing coordinate.
        resolution (int): The resolution expressed as a metre-based integer.
        expected (str): The expected result is a BNG reference formatted string.
    """

    easting: int | float
    northing: int | float
    resolution: int
    expected: str


# Parameterised test for _get_bng_suffix function
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "indexing_test_cases.json"
    )["_get_bng_suffix"],
)
def test__get_bng_suffix(test_case: GetBNGSuffixTestCase) -> None:
    """Test _get_bng_suffix function with test cases from JSON file.

    Args:
        test_case (GetBNGSuffixTestCase): Test case from JSON file.
    """
    # Load test case data
    easting = test_case["easting"]
    northing = test_case["northing"]
    resolution = test_case["resolution"]
    expected = test_case["expected"]

    # Assert that the function returns the expected result
    assert _get_bng_suffix(easting, northing, resolution) == expected


class DecomposeGeomTestCase(TypedDict):
    """TypedDict for _decompose_geom function test cases.

    Attributes:
        geom (dict[str, Any]): Geometry reresented in GeoJSON format.
        expected (dict[str, int | list[str]]): Expected result is a dictionary with the
            expected part count and list of part geometry types.
    """

    geom: dict[str, Any]
    expected: dict[str, int | list[str]]


# Parameterised test for _decompose_geom function
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "indexing_test_cases.json"
    )["_decompose_geom"],
)
def test__decompose_geom(test_case: DecomposeGeomTestCase) -> None:
    """Test _decompose_geom with test cases from JSON file.

    Args:
        test_case (DecomposeGeomTestCase): Test case from JSON file.
    """
    # Load test case data
    # Convert test case geometry from GeoJSON to Shapely Geometry object
    geom = shape(test_case["geom"])
    expected_count = test_case["expected"]["count"]
    expected_types = test_case["expected"]["types"]

    # Decompose geometry into its constituent parts
    parts = _decompose_geom(geom)

    # Assert that the decomposition returns the expected part count
    assert len(parts) == expected_count
    # Assert that the decomposition returns the expected part types
    types = [part.geom_type for part in parts]
    assert sorted(types) == sorted(expected_types)


class XYToBNGTestCase(TypedDict):
    """TypedDict for xy_to_bng function test cases.

    Attributes:
        easting (int | float): The easting coordinate.
        northing (int | float): The northing coordinate.
        resolution (int | str): The resolution expressed either as a metre-based
            integer or as a string label.
        expected (dict[str, str] | None): The expected result is a dictionary with the
            BNG reference formatted string.
        expected_exception (dict[str, str] | None): The expected exception is a
            dictionary with the exception name.
    """

    easting: int | float
    northing: int | float
    resolution: int | str
    expected: dict[str, str] | None
    expected_exception: dict[str, str] | None


# Parameterised test for xy_to_bng function
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "indexing_test_cases.json"
    )["xy_to_bng"],
)
def test_xy_to_bng(test_case: XYToBNGTestCase) -> None:
    """Test xy_to_bng with test cases from JSON file.

    Args:
        test_case (XYToBNGTestCase): Test case from JSON file.
    """
    # Load test case data
    easting = test_case["easting"]
    northing = test_case["northing"]
    resolution = test_case["resolution"]

    if "expected_exception" in test_case:
        # Get exception name from test case
        exception_name = test_case["expected_exception"]["name"]
        # Get exception class from name
        exception_class = _EXCEPTION_MAP[exception_name]
        # Assert that the test case raises the expected exception
        with pytest.raises(exception_class):
            xy_to_bng(easting, northing, resolution)

    else:
        # Get expected result
        expected = test_case["expected"]["bng_ref_formatted"]
        # Create BNGReference object
        bng_ref = xy_to_bng(easting, northing, resolution)
        # Assert that the function returns the expected result
        assert bng_ref.bng_ref_formatted == expected


class BNGToXYTestCase(TypedDict):
    """TypedDict for bng_to_xy function test cases.

    Attributes:
        bng_ref_string (str): The BNG reference string.
        position (str): The grid cell position expressed as a string.
        expected (list[int | float]): The expected result is a list of easting and
            northing coordinates.
    """

    bng_ref_string: str
    position: int | str
    expected: list[int | float]


# Parameterised test for bng_to_xy function
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "indexing_test_cases.json"
    )["bng_to_xy"],
)
def test_bng_to_xy(test_case: BNGToXYTestCase) -> None:
    """Test bng_to_xy with test cases from JSON file.

    Args:
        test_case (BNGToXYTestCase): Test case from JSON file.
    """
    # Load test case data
    bng_ref_string = test_case["bng_ref_string"]
    position = test_case["position"]
    # Get expected result as tuple
    expected = tuple(test_case["expected"])

    # Create BNGReference object
    bng_ref = BNGReference(bng_ref_string)

    # Assert that the function returns the expected result
    assert bng_to_xy(bng_ref, position=position) == expected


class BNGToBBOXTestCase(TypedDict):
    """TypedDict for bng_to_bbox function test cases.

    Attributes:
        bng_ref_string (str): The BNG reference string.
        expected (list[int]): The expected result is a list of bounding box coordinates.
    """

    bng_ref_string: str
    expected: list[int]


# Parameterised test for bng_to_bbox function
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "indexing_test_cases.json"
    )["bng_to_bbox"],
)
def test_bng_to_bbox(test_case: BNGToBBOXTestCase) -> None:
    """Test bng_to_bbox with test cases from JSON file.

    Args:
        test_case (BNGToBBOXTestCase): Test case from JSON file.
    """
    # Load test case data
    bng_ref_string = test_case["bng_ref_string"]
    # Get expected result as tuple
    expected = tuple(test_case["expected"])

    # Create BNGReference object
    bng_ref = BNGReference(bng_ref_string)

    # Assert that the function returns the expected result
    assert bng_to_bbox(bng_ref) == expected


class BNGToGridGeomTestCase(TypedDict):
    """TypedDict for bng_to_grid_geom function test cases.

    Attributes:
        bng_ref_string (str): The BNG reference string.
        expected (dict[str, Any]): The expected result is a dictionary with the
            expected geometry in GeoJSON format.
    """

    bng_ref_string: str
    expected: dict[str, Any]


# Parameterised test for bng_to_grid_geom function
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "indexing_test_cases.json"
    )["bng_to_grid_geom"],
)
def test_bng_to_grid_geom(test_case: BNGToGridGeomTestCase) -> None:
    """Test bng_to_grid_geom with test cases from JSON file.

    Args:
        test_case (BNGToGridGeomTestCase): Test case from JSON file.
    """
    # Load test case data
    bng_ref_string = test_case["bng_ref_string"]
    # Convert expected result from GeoJSON to Shapely geometry object
    expected = shape(test_case["expected"])

    # Create BNGReference object
    bng_ref = BNGReference(bng_ref_string)

    # Assert that the the two geometries are equal
    # Normalise geometries to account for coordinate order differences
    assert_geometries_equal(bng_to_grid_geom(bng_ref), expected, normalize=True)


class BBOXToBNGTestCase(TypedDict):
    """TypedDict for bbox_to_bng function test cases.

    Attributes:
        xmin (int | float): The minimum easting coordinate of the bounding box (BBOX).
        ymin (int | float): The minimum northing coordinate of the BBOX.
        xmax (int | float): The maximum easting coordinate of the BBOX.
        ymax (int | float): The maximum northing coordinate of the BBOX.
        resolution (int | str): The resolution expressed either as a metre-based
            integer or as a string label.
        expected_warning (bool | None): The expected warning is a boolean indicating if
            a warning is expected.
        expected (dict[str, list[str]]): The expected result is a dictionary with a
            list of BNG reference formatted strings.
    """

    xmin: int | float
    ymin: int | float
    xmax: int | float
    ymax: int | float
    resolution: int | str
    expected_warning: bool | None
    expected: dict[str, list[str]]


# Parameterised test for bbox_to_bng function
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "indexing_test_cases.json"
    )["bbox_to_bng"],
)
def test_bbox_to_bng(test_case: BBOXToBNGTestCase) -> None:
    """Test bbox_to_bng with test cases from JSON file.

    Args:
        test_case (BBOXToBNGTestCase): Test case from JSON file.
    """
    # Load test case data
    xmin = test_case["xmin"]
    ymin = test_case["ymin"]
    xmax = test_case["xmax"]
    ymax = test_case["ymax"]
    resolution = test_case["resolution"]
    # Get expected result
    expected = test_case["expected"]["bng_ref_formatted"]

    if "expected_warning" in test_case:
        # Assert that the function raises a warning
        with pytest.warns(UserWarning):
            # Return a list of BNGReference objects
            bng_refs = bbox_to_bng(xmin, ymin, xmax, ymax, resolution)
            # Sort lists to account for order differences
            bng_ref_strings = [bng_ref.bng_ref_formatted for bng_ref in bng_refs]
            # Assert that the function returns the expected result
            assert sorted(bng_ref_strings) == sorted(expected)

    else:
        # Return a list of BNGReference objects
        bng_refs = bbox_to_bng(xmin, ymin, xmax, ymax, resolution)
        # Assert that the function returns the expected result
        # Sort lists to account for order differences
        bng_ref_strings = [bng_ref.bng_ref_formatted for bng_ref in bng_refs]
        # Assert that the function returns the expected result
        assert sorted(bng_ref_strings) == sorted(expected)


class GeomToBNGTestCase(TypedDict):
    """TypedDict for geom_to_bng function test cases.

    Attributes:
        geom (dict[str, Any]): Geometry represented in GeoJSON format.
        resolution (int | str): The resolution expressed either as a metre-based
            integer or as a string label.
        expected_exception (dict[str, str] | None): The expected exception is a
            dictionary with the exception name.
        expected_warning (bool | None): The expected warning is a boolean indicating if
            a warning is expected.
        expected (dict[str, list[str]] | None): The expected result is a dictionary
            with a list of BNG reference formatted strings.
    """

    geom: dict[str, Any]
    resolution: int | str
    expected_exception: dict[str, str] | None
    expected_warning: bool | None
    expected: dict[str, list[str]] | None


def validate_and_assert_bng_intersects(
    geom: Geometry, resolution: int | str, expected: list[str]
) -> None:
    """Helper function to validate and assert geom_to_bng function return.

    Args:
        geom (Geometry): Shapely Geometry object.
        resolution (int | str): The resolution expressed either as a metre-based
            integer or as a string label.
        expected (list[str]): Expected result. A list containing the expected BNG
            reference formatted strings.
    """
    # Return a list of BNGReference objects
    bng_refs = geom_to_bng(geom, resolution)
    # Sort lists to account for order differences
    bng_ref_strings = [bng_ref.bng_ref_formatted for bng_ref in bng_refs]
    # Assert that the function returns the expected result
    assert sorted(bng_ref_strings) == sorted(expected)


# Parameterised test for geom_to_bng function
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "indexing_test_cases.json"
    )["geom_to_bng"],
)
def test_geom_to_bng(test_case: GeomToBNGTestCase) -> None:
    """Test geom_to_bng with test cases from JSON file.

    Args:
        test_case (GeomToBNGTestCase): Test case from JSON file.
    """
    # Load test case data
    # Convert test case geometry from GeoJSON to Shapely Geometry object
    geom = shape(test_case["geom"])
    resolution = test_case["resolution"]
    # Get expected result
    expected = (
        None
        if "expected_exception" in test_case
        else test_case["expected"]["bng_ref_formatted"]
    )

    if "expected_exception" in test_case:
        # Get exception name from test case
        exception_name = test_case["expected_exception"]["name"]
        # Get exception class from name
        exception_class = _EXCEPTION_MAP[exception_name]
        # Assert that the test case raises the expected exception
        with pytest.raises(exception_class):
            geom_to_bng(geom, resolution)

    elif "expected_warning" in test_case:
        # Assert that the function raises a warning
        with pytest.warns(UserWarning):
            # Assert that the function returns the expected result
            validate_and_assert_bng_intersects(geom, resolution, expected)

    else:
        # Assert that the function returns the expected result
        validate_and_assert_bng_intersects(geom, resolution, expected)


class GeomToBNGIntersectionTestCase(TypedDict):
    """TypedDict for geom_to_bng_intersection function test cases.

    Attributes:
        geom (dict[str, Any]): Geometry represented in GeoJSON format.
        resolution (int | str): The resolution expressed either as a metre-based
            integer or as a string label.
        expected_exception (dict[str, str] | None): The expected exception is a
            dictionary with the exception name.
        expected_warning (bool | None): The expected warning is a boolean indicating if
            a warning is expected.
        expected (list[dict[str, str | bool]] | None): The expected result is a list of
            dictionaries with the expected BNG reference formatted strings and booleans
            indicating if a grid square is a core geometry.
    """

    geom: dict[str, Any]
    resolution: int | str
    expected_exception: dict[str, str] | None
    expected_warning: bool | None
    expected: list[dict[str, str | bool]] | None


def validate_and_assert_bng_intersection(
    geom: Geometry, resolution: int | str, expected: dict[str, str | bool]
) -> None:
    """Helper function to validate and assert geom_to_bng_intersection return.

    Args:
        geom (Geometry): Shapely Geometry object.
        resolution (int | str): The resolution expressed either as a metre-based
            integer or as a string label.
        expected (dict[str, str | bool]): Expected result. A dictionary containing the
            expected BNG reference formatted string and a boolean indicating if it is a
            core geometry.
    """
    # Convert test case geometry from GeoJSON to Shapely Geometry object
    # Return a list of BNGIndexedGeometry objects
    bng_idx_geoms = geom_to_bng_intersection(geom, resolution)
    # Extract bng_ref_formatted and is_core properties to create a simplified
    # representation of the BNGIndexedGeometry objects for comparison with the
    # expected output.
    result = [
        (bng_idx_geom.bng_ref.bng_ref_formatted, bng_idx_geom.is_core)
        for bng_idx_geom in bng_idx_geoms
    ]
    # Assert that the result matches the expected output
    assert sorted(result) == sorted(expected)
    # Extract the areas of the core indexed geometries
    # Core indexed geometries represent grid squares that are
    # fully contained within the input geometry
    result_core_areas = [
        bng_idx_geom.geom.area for bng_idx_geom in bng_idx_geoms if bng_idx_geom.is_core
    ]
    if result_core_areas:
        # Normalise the resolution to its metre equivalent
        normalised_resolution = _validate_and_normalise_bng_resolution(resolution)
        # Assert that the resolution of the core indexed geometries
        # is equal to the normalised resolution
        assert all(sqrt(area) == normalised_resolution for area in result_core_areas)


# Parameterised test for geom_to_bng_intersection function
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "indexing_test_cases.json"
    )["geom_to_bng_intersection"],
)
def test_geom_to_bng_intersection(test_case: GeomToBNGIntersectionTestCase) -> None:
    """Test geom_to_bng_intersection with test cases from JSON file.

    Args:
        test_case (GeomToBNGIntersectionTestCase): Test case from JSON file.
    """
    # Load test case data
    # Convert test case geometry from GeoJSON to Shapely Geometry object
    geom = shape(test_case["geom"])
    resolution = test_case["resolution"]
    # Convert expected result dictionary values into tuples
    expected = (
        None
        if "expected_exception" in test_case
        else [
            (item["bng_ref_formatted"], item["is_core"])
            for item in test_case["expected"]
        ]
    )

    if "expected_exception" in test_case:
        # Get exception name from test case
        exception_name = test_case["expected_exception"]["name"]
        # Get exception class from name
        exception_class = _EXCEPTION_MAP[exception_name]
        # Assert that the test case raises an exception
        with pytest.raises(exception_class):
            geom_to_bng_intersection(geom, resolution)

    elif "expected_warning" in test_case:
        # Assert that the test case raises a warning
        with pytest.warns(UserWarning):
            # Assert that the function returns the expected result
            validate_and_assert_bng_intersection(geom, resolution, expected)

    else:
        # Assert that the function returns the expected result
        validate_and_assert_bng_intersection(geom, resolution, expected)



"""Testing for the indexing_gpd module.

Testing is skipped if the GeoPandas (https://github.com/geopandas/geopandas) package is
not installed.

The test cases are defined in the JSON file located at ./data/indexing_test_cases.json
and are used to parameterise the tests for various functions in the indexing module.
Test cases are loaded from the JSON file using the _load_test_cases function, which is
defined in the utils module. The test cases are defined as TypedDicts, which provide a
way to define the structure of the test case data.

Testing reuses the test cases for the osbng.indexing.geom_to_bng_intersection function.
"""

from math import sqrt
from pathlib import Path

import pytest

# Skip tests if geopandas package is not installed
# Match minimum geopandas version to (optional) dependency in pyproject.toml
pytest.importorskip(
    "geopandas",
    minversion="1.0.0",
    reason="The 'geopandas' package is required for 'indexing_gpd' tests.",
)

import geopandas as gpd
from shapely.geometry import shape

from osbng.errors import _EXCEPTION_MAP
from osbng.indexing import _validate_and_normalise_bng_resolution
from osbng.indexing_gpd import gdf_to_bng_intersection_explode
from osbng.utils import _load_test_cases
from tests.test_indexing import GeomToBNGIntersectionTestCase


def validate_and_assert_gdf_bng_intersection(
    gdf: gpd.GeoDataFrame, resolution: int | str, expected: list[tuple[str, bool]]
) -> None:
    """Validates and asserts gdf_to_bng_intersection_explode return.

    Args:
        gdf (gpd.GeoDataFrame): GeoPandas GeoDataFrame containing the geometry to be
            tested.
        resolution (int | str): The resolution expressed either as a metre-based
            integer or as a string label.
        expected (list[tuple[str, bool]]): Expected result. A list of tuples, where
            each tuple contains, the expected BNG reference formatted string and a
            boolean indicating if it is a core geometry.
    """
    # Apply the gdf_to_bng_intersection_explode function to the input GeoDataFrame
    gdf_test = gdf_to_bng_intersection_explode(gdf, resolution)
    # Assert that the result is a GeoDataFrame
    assert isinstance(gdf_test, gpd.GeoDataFrame)

    # Extract the 'bng_ref_formatted' property from the 'bng_ref' column
    # and 'is_core' column
    result = [
        (bng_ref.bng_ref_formatted, is_core)
        for bng_ref, is_core in zip(gdf_test["bng_ref"], gdf_test["is_core"])
    ]
    # Assert that the result matches the expected output
    assert sorted(result) == sorted(expected)

    # Extract the areas of the core indexed geometries
    # Core indexed geometries represent grid squares that are fully contained within
    # the input geometry
    result_core_areas = gdf_test[gdf_test["is_core"]]["geometry"].area.tolist()

    if result_core_areas:
        # Normalise the resolution to its metre equivalent
        normalised_resolution = _validate_and_normalise_bng_resolution(resolution)
        # Assert that the resolution of the core indexed geometries
        # is equal to the normalised resolution
        assert all(sqrt(area) == normalised_resolution for area in result_core_areas)


# Parameterised test for gdf_to_bng_intersection_explode function
@pytest.mark.parametrize(
    "test_case",
    # Load test cases from JSON file
    # Reuses test cases for the osbng.indexing.geom_to_bng_intersection function
    _load_test_cases(
        file_path=Path(__file__).parent / "data" / "indexing_test_cases.json"
    )["geom_to_bng_intersection"],
)
def test_gdf_to_bng_intersection_explode(
    test_case: GeomToBNGIntersectionTestCase,
) -> None:
    """Test gdf_to_bng_intersection_explode with test cases from JSON file.

    Args:
        test_case (GeomToBNGIntersectionTestCase): Test case from JSON file.
    """
    # Load test case data
    # Convert test case geometry from GeoJSON to Shapely Geometry object
    geom = shape(test_case["geom"])
    # Create GeoDataFrame from the geometry
    # Set GeoDataFrame coordinate reference system (CRS) to
    # 'EPSG:27700' (British National Grid)
    gdf = gpd.GeoDataFrame({"geometry": [geom]}, crs=27700)
    resolution = test_case["resolution"]
    # Convert expected result dictionary values into tuples
    expected = (
        None
        if "expected_exception" in test_case
        else [
            (item["bng_ref_formatted"], item["is_core"])
            for item in test_case["expected"]
        ]
    )

    if "expected_exception" in test_case:
        # Get exception name from test case
        exception_name = test_case["expected_exception"]["name"]
        # Get exception class from name
        exception_class = _EXCEPTION_MAP[exception_name]
        # Assert that the test case raises an exception
        with pytest.raises(exception_class):
            gdf_to_bng_intersection_explode(gdf, resolution)

    elif "expected_warning" in test_case:
        # Assert that the test case raises a warning
        with pytest.warns(UserWarning):
            # Assert that the function returns the expected result
            validate_and_assert_gdf_bng_intersection(gdf, resolution, expected)

    else:
        # Assert that the function returns the expected result
        validate_and_assert_gdf_bng_intersection(gdf, resolution, expected)



"""Support for parameterised testing."""

import json
from pathlib import Path


def _load_test_cases(file_path: str | Path) -> dict:
    """Load test cases from a JSON file.

    Args:
        file_path (str | Path): Path to the JSON file containing test cases.

    Returns:
        dict: Test cases as a dictionary.

    """
    # Convert to Path object if a string is provided
    p = Path(file_path)
    with p.open(encoding="utf-8") as f:
        return json.load(f)
